import json
import re
import sys
import unittest
from pathlib import Path
from unittest.mock import patch
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import requests
import responses

try:
    from onelogin.saml2.utils import OneLogin_Saml2_Utils

    SAML_MODULE_ENABLED = True
except ImportError:
    SAML_MODULE_ENABLED = False

from social_core.exceptions import AuthMissingParameter

from .base import BaseBackendTest

DATA_DIR = Path(__file__).parent / "data"


@unittest.skipIf(
    "__pypy__" in sys.builtin_module_names, "dm.xmlsec not compatible with pypy"
)
@unittest.skipUnless(SAML_MODULE_ENABLED, "Only run if onelogin.saml2 is installed")
class SAMLTest(BaseBackendTest):
    backend_path = "social_core.backends.saml.SAMLAuth"
    expected_username = "myself"
    response_fixture = "saml_response.txt"

    def extra_settings(self):
        file = DATA_DIR / "saml_config.json"
        config_str = file.read_text()
        return json.loads(config_str)

    def setUp(self) -> None:
        """Patch the time so that we can replay canned
        request/response pairs"""
        super().setUp()

        @staticmethod
        def fixed_time():
            return OneLogin_Saml2_Utils.parse_SAML_to_time("2015-05-09T03:57:22Z")

        now_patch = patch.object(OneLogin_Saml2_Utils, "now", fixed_time)
        now_patch.start()
        self.addCleanup(now_patch.stop)

    def install_http_intercepts(self, start_url, return_url) -> None:
        # When we request start_url
        # (https://idp.testshib.org/idp/profile/SAML2/Redirect/SSO...)
        # we will eventually get a redirect back, with SAML assertion
        # data in the query string.  A pre-recorded correct response
        # is kept in this .txt file:
        file = DATA_DIR / self.response_fixture
        response_url = file.read_text()
        responses.add(
            responses.GET,
            start_url,
            status=301,
            headers={"Location": response_url},
        )
        responses.add(responses.GET, return_url, status=200, body="foobar")

    def do_start(self):
        start_url = self.backend.start().url
        # Modify the start URL to make the SAML request consistent
        # from test to test:
        start_url = self.modify_start_url(start_url)
        # If the SAML Identity Provider recognizes the user, we will
        # be redirected back to:
        return_url = self.backend.redirect_uri
        self.install_http_intercepts(start_url, return_url)
        response = requests.get(start_url, timeout=1)
        self.assertTrue(response.url.startswith(return_url))
        self.assertEqual(response.text, "foobar")
        query_values = {
            k: v[0] for k, v in parse_qs(urlparse(response.url).query).items()
        }
        self.assertNotIn(" ", query_values["SAMLResponse"])
        self.strategy.set_request_data(query_values, self.backend)
        return self.backend.complete()

    def test_metadata_generation(self) -> None:
        """Test that we can generate the metadata without error"""
        xml, errors = self.backend.generate_metadata_xml()
        self.assertEqual(len(errors), 0)
        self.assertEqual(xml.decode()[0], "<")

    def test_login_with_next_url(self) -> None:
        """
        Test that we login and then redirect to the "next" URL.
        """
        # pretend we've started with a URL like /login/saml/?idp=testshib&next=/foo/bar
        self.strategy.set_request_data(
            {"idp": "testshib", "next": "/foo/bar"}, self.backend
        )
        self.do_login()
        # The core `do_complete` action assumes the "next" URL is stored in session state or the request data.
        self.assertEqual(self.strategy.session_get("next"), "/foo/bar")

    def test_login_no_next_url(self) -> None:
        """
        Test that we handle "next" being omitted from the request data and RelayState.
        """
        self.response_fixture = "saml_response_no_next_url.txt"

        # pretend we've started with a URL like /login/saml/?idp=testshib
        self.strategy.set_request_data({"idp": "testshib"}, self.backend)
        self.do_login()
        self.assertEqual(self.strategy.session_get("next"), None)

    def test_login_with_legacy_relay_state(self) -> None:
        """
        Test that we handle legacy RelayState (i.e. just the IDP name, not a JSON object).

        This is the form that RelayState had in prior versions of this library. It should be supported for backwards
        compatibility. It is also sent by some identity providers (like Okta) on IdP-initiated login.
        """
        self.response_fixture = "saml_response_legacy.txt"

        self.strategy.set_request_data({"idp": "testshib"}, self.backend)
        self.do_login()

    def test_login_no_idp_in_initial_request(self) -> None:
        """
        Logging in without an idp param should raise AuthMissingParameter
        """
        with self.assertRaises(AuthMissingParameter):
            self.do_start()

    def test_login_no_idp_in_saml_response(self) -> None:
        """
        The RelayState should always contain a JSON object with an "idp" key, or be just the IDP name as a string.
        This tests that an exception is raised if it is a JSON object, but is missing the "idp" key.
        """
        self.response_fixture = "saml_response_no_idp_name.txt"

        with self.assertRaises(AuthMissingParameter):
            self.do_start()

    def modify_start_url(self, start_url):
        """
        Given a SAML redirect URL, parse it and change the ID to
        a consistent value, so the request is always identical.
        """
        # Parse the SAML Request URL to get the XML being sent to TestShib
        url_parts = urlparse(start_url)
        query = {k: v[0] for (k, v) in parse_qs(url_parts.query).items()}
        xml = OneLogin_Saml2_Utils.decode_base64_and_inflate(query["SAMLRequest"])
        # Modify the XML:
        xml = xml.decode()
        xml, changed = re.subn(r'ID="[^"]+"', 'ID="TEST_ID"', xml)
        self.assertEqual(changed, 1)
        # Update the URL to use the modified query string:
        query["SAMLRequest"] = OneLogin_Saml2_Utils.deflate_and_base64_encode(xml)
        url_parts = list(url_parts)
        url_parts[4] = urlencode(query)
        return urlunparse(url_parts)



"""
Pocket OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/pocket.html
"""

from __future__ import annotations

from typing import Any

from social_core.utils import handle_http_errors

from .base import BaseAuth


class PocketAuth(BaseAuth):
    name = "pocket"
    AUTHORIZATION_URL = "https://getpocket.com/auth/authorize"
    ACCESS_TOKEN_URL = "https://getpocket.com/v3/oauth/authorize"
    REQUEST_TOKEN_URL = "https://getpocket.com/v3/oauth/request"
    ID_KEY = "username"

    def get_json(self, url, *args, **kwargs):
        headers = {"X-Accept": "application/json"}
        kwargs.update({"method": "POST", "headers": headers})
        return super().get_json(url, *args, **kwargs)

    def get_user_details(self, response):
        return {"username": response["username"]}

    def extra_data(
        self,
        user,
        uid: str,
        response: dict[str, Any],
        details: dict[str, Any],
        pipeline_kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        return response

    def auth_url(self) -> str:
        data = {
            "consumer_key": self.setting("KEY"),
            "redirect_uri": self.redirect_uri,
        }
        token = self.get_json(self.REQUEST_TOKEN_URL, data=data)["code"]
        self.strategy.session_set("pocket_request_token", token)
        return f"{self.AUTHORIZATION_URL}?request_token={token}&redirect_uri={self.redirect_uri}"

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        data = {
            "consumer_key": self.setting("KEY"),
            "code": self.strategy.session_get("pocket_request_token"),
        }
        response = self.get_json(self.ACCESS_TOKEN_URL, data=data)
        kwargs.update({"response": response, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class YahooOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.yahoo.YahooOAuth2"
    user_data_url = "https://api.login.yahoo.com/openid/v1/userinfo"
    expected_username = "j.doe"
    access_token_body = json.dumps(
        {
            "access_token": "UNQO1djO5xpaKm3_KbECBKB5mlFr6tSZTOLrrJCprtT1X1UFljpxiS5iSue8u_n8ah1WbL6sTNw3HPFHicyXDbTs7aSrbIe.rx9n9dzX7xZjx8dyF2Ap1a6J_nw4k56a5mCOuTd.ZFQENgGtHwM0DRFVeDNTAx_WzhqDGPCqhtsNICuuY30soFZGS11FTlUk7Gy0ISjxLRAjIZVtpojnY5p8XuT1qUtAheWqZegJ_7t.AP4o0J4xJ3_oocXeiSKEXaD3AijdBdViKPZI3Ow7yeHK8uX1weNfKoSP6eEpCviyj0YlRMIBSg4cRdGL6EsSggX6B5gzgcA9efDSpcwVhupY0RlUdi.AxJ1nT0frWmrYiwntpu1XP_5mIbOlb4wfrD_ZCRNY2Qby40RBt5iHERSJ89K1o69fw3Jd4C3hF14iJLHcDHmnYJSX651G9MlpGPWT99DRteCdhSm8URbZqfGPG8mZtLpmhfxr1umCoGEgocrfHpITMjOyEwvgmAhgjGKXugvdNTABn0AEQBetIVtJ80Ymbn6IMq_Qh10vyspVsVK69C9yTlwLtZhcvim5Nk_15JHd0GSj0Mj.X.FWTzUK1e3CNQjeJxdQ2Qk9BXDC4_DXW_Ot5LzYy5qRvRKT4gh54n5aBROxFdky0ELt1IgkLTRJ0idUCen87klP.0CLp1QTNXx99N6nM9c_HwWVKwhILUjzXaIrP0GVEMwlGIHqn2I91Z03irBgzrMB219lqUAuF27_OD4QnyQfICSW65n5hVo1e89xwN6VN3usRrhHmdDfd7nk3nzMyXdsOPzghA1huBCYyEGZ_kq9FzVFQ5QYDmJ0WqpmG1yXDEntYVvkB_i_jkbNPH4.R134ptwznCZSuQ--",
            "refresh_token": "AJj.Dlbt_e4XN85buQhFXj77sIB3lqBF3Bcqb2kwUEoYrBb0Pg--",
            "expires_in": 3600,
            "token_type": "bearer",
            "xoauth_yahoo_guid": "UQIDWJNWVNQD4GXZ5NGMZUSTQ4",
            "id_token": "eyJhbGciOiJFUzI1NiIsImtpZCI6IjM0NjZkNTFmN2RkMGM3ODA1NjU2ODhjMTgzOTIxODE2YzQ1ODg5YWQifQ.eyJhdF9oYXNoIjoiYWM5YkR3ejVMWjl5UEVpdWtEcGdzdz09Iiwic3ViIjoiVVFJRFdKTldWTlFENEdYWjVOR01aVVNUUTQiLCJhdWQiOiJkajB5Sm1rOVdHeDBRbEUwVVdkQ2EwaEtKbVE5V1Zkck9XTnJOVWhYVm5CaFRraEZiV05IYnpsTlFTMHRKbk05WTI5dWMzVnRaWEp6WldOeVpYUW1lRDAxT0EtLSIsImlzcyI6Imh0dHBzOi8vbG9naW4ueWFob28uY29tIiwiZXhwIjoxNDQzODI3MTMwLCJub25jZSI6IjEyMzQ1IiwiaWF0IjoxNDQzODIzNTMwfQ.n7oEFi5028StcI41Hkh6lLYK4PmF7pT4AIXrQ_62nfDEZj2g0oYjSLFPJp4IqF6LefwcCQ9FHT5X9eC8A7peqw",
        }
    )
    user_data_body = json.dumps(
        {
            "sub": "FSVIDUW3D7FSVIDUW3D72F2F",  # user identifier
            "name": "Jane Doe",
            "given_name": "Jane",
            "family_name": "Doe",
            "preferred_username": "j.doe",
            "email": "janedoe@example.com",
            "picture": "http://example.com/janedoe/me.jpg",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Copyright (c) 2015 Microsoft Open Technologies, Inc.

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Azure AD OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/azuread.html

See https://nicksnettravels.builttoroam.com/post/2017/01/24/Verifying-Azure-Active-Directory-JWT-Tokens.aspx
for verifying JWT tokens.
"""

from __future__ import annotations

import base64
from typing import TYPE_CHECKING, Any, cast

from cryptography.x509 import load_der_x509_certificate
from jwt import DecodeError, ExpiredSignatureError, get_unverified_header
from jwt import decode as jwt_decode

from social_core.exceptions import AuthTokenError

from .azuread import AzureADOAuth2

if TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
    from cryptography.x509 import Certificate


class AzureADTenantOAuth2(AzureADOAuth2):
    name = "azuread-tenant-oauth2"
    OPENID_CONFIGURATION_URL = "{base_url}/.well-known/openid-configuration{appid}"
    JWKS_URL = "{base_url}/discovery/keys{appid}"

    @property
    def tenant_id(self) -> str:
        return cast("str", self.setting("TENANT_ID", "common"))

    def openid_configuration_url(self):
        return self.OPENID_CONFIGURATION_URL.format(
            base_url=self.base_url, appid=self._appid()
        )

    def jwks_url(self):
        return self.JWKS_URL.format(base_url=self.base_url, appid=self._appid())

    def _appid(self) -> str:
        return (
            f"?appid={self.setting('KEY')}" if self.setting("KEY") is not None else ""
        )

    def get_certificate(self, kid: str) -> Certificate:
        # retrieve keys from jwks_url
        resp = self.request(self.jwks_url(), method="GET")
        resp.raise_for_status()

        # find the proper key for the kid
        for key in resp.json()["keys"]:
            if key["kid"] == kid:
                x5c = key["x5c"][0]
                break
        else:
            raise DecodeError(f"Cannot find kid={kid}")

        return load_der_x509_certificate(base64.b64decode(x5c))

    def get_user_id(self, details, response):
        """Use subject (sub) claim as unique id."""
        return response.get("sub")

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        response = kwargs.get("response")
        if response and response.get("id_token"):
            id_token = response.get("id_token")
        else:
            id_token = access_token

        # get key id and algorithm
        key_id = get_unverified_header(id_token)["kid"]

        try:
            # retrieve certificate for key_id
            certificate = self.get_certificate(key_id)

            return jwt_decode(
                id_token,
                key=cast("RSAPublicKey", certificate.public_key()),
                algorithms=["RS256"],
                audience=self.setting("KEY"),
            )
        except (DecodeError, ExpiredSignatureError) as error:
            raise AuthTokenError(self, error) from error


class AzureADV2TenantOAuth2(AzureADTenantOAuth2):
    name = "azuread-v2-tenant-oauth2"
    OPENID_CONFIGURATION_URL = "{base_url}/v2.0/.well-known/openid-configuration{appid}"
    AUTHORIZATION_URL = "{base_url}/oauth2/v2.0/authorize"
    ACCESS_TOKEN_URL = "{base_url}/oauth2/v2.0/token"
    JWKS_URL = "{base_url}/discovery/v2.0/keys{appid}"
    DEFAULT_SCOPE = ["openid", "profile", "offline_access"]

    def get_user_id(self, details, response):
        """Use upn as unique id"""
        return response.get("preferred_username")

    def get_user_details(self, response):
        """Return user details from Azure AD account"""
        fullname, first_name, last_name = (
            response.get("name", ""),
            response.get("given_name", ""),
            response.get("family_name", ""),
        )
        return {
            "username": fullname,
            "email": response.get("preferred_username"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class ReadabilityOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.readability.ReadabilityOAuth"
    user_data_url = "https://www.readability.com/api/rest/v1/users/_current"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps(
        {
            "username": "foobar",
            "first_name": "Foo",
            "last_name": "Bar",
            "has_active_subscription": False,
            "tags": [],
            "is_publisher": False,
            "email_into_address": "foobar+sharp@inbox.readability.com",
            "kindle_email_address": None,
            "avatar_url": "https://secure.gravatar.com/avatar/"
            "5280f15cedf540b544eecc30fcf3027c?d="
            "https://www.readability.com/media/images/"
            "avatar.png&s=36",
            "date_joined": "2013-03-18 02:51:02",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Facebook OAuth2, and Canvas Application  backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/facebook.html
"""

import base64
import hashlib
import hmac
import json
import time
from typing import Any, cast

from social_core.exceptions import (
    AuthCanceled,
    AuthException,
    AuthMissingParameter,
    AuthUnknownError,
)
from social_core.utils import constant_time_compare, handle_http_errors, parse_qs

from .oauth import BaseOAuth2

API_VERSION = 24.0


class FacebookOAuth2(BaseOAuth2):
    """Facebook OAuth2 authentication backend"""

    name = "facebook"
    REDIRECT_STATE = False
    RESPONSE_TYPE = None
    SCOPE_SEPARATOR = ","
    AUTHORIZATION_URL = "https://www.facebook.com/v{version}/dialog/oauth"
    ACCESS_TOKEN_URL = "https://graph.facebook.com/v{version}/oauth/access_token"
    ACCESS_TOKEN_METHOD = "GET"
    REVOKE_TOKEN_URL = "https://graph.facebook.com/v{version}/{uid}/permissions"
    REVOKE_TOKEN_METHOD = "DELETE"
    USER_DATA_URL = "https://graph.facebook.com/v{version}/me"
    EXTRA_DATA = [
        ("id", "id"),
        ("expires_in", "expires_in"),
        ("granted_scopes", "granted_scopes"),
        ("denied_scopes", "denied_scopes"),
    ]

    def auth_params(self, state=None):
        params = super().auth_params(state)
        params["return_scopes"] = "true"
        return params

    def get_authorization_url_format(self) -> dict[str, str]:
        return {"version": cast("str", self.setting("API_VERSION", API_VERSION))}

    def get_access_token_url_format(self) -> dict[str, str]:
        return {"version": cast("str", self.setting("API_VERSION", API_VERSION))}

    def get_user_details(self, response):
        """Return user details from Facebook account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("name", ""),
            response.get("first_name", ""),
            response.get("last_name", ""),
        )
        return {
            "username": response.get("username", response.get("name")),
            "email": response.get("email", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        params = self.setting("PROFILE_EXTRA_PARAMS", {}).copy()
        params["access_token"] = access_token

        if self.setting("APPSECRET_PROOF", True):
            _, secret = self.get_key_and_secret()
            params["appsecret_proof"] = hmac.new(
                secret.encode("utf8"),
                msg=access_token.encode("utf8"),
                digestmod=hashlib.sha256,
            ).hexdigest()

        version = self.setting("API_VERSION", API_VERSION)
        return self.get_json(self.USER_DATA_URL.format(version=version), params=params)

    def process_error(self, data) -> None:
        super().process_error(data)
        if data.get("error_code"):
            raise AuthCanceled(
                self, data.get("error_message") or data.get("error_code")
            )

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        if not self.data.get("code"):
            raise AuthMissingParameter(self, "code")
        state = self.validate_state()
        key, secret = self.get_key_and_secret()
        response = self.request(
            self.access_token_url(),
            params={
                "client_id": key,
                "redirect_uri": self.get_redirect_uri(state),
                "client_secret": secret,
                "code": self.data["code"],
            },
            method=self.ACCESS_TOKEN_METHOD,
        )
        # API v2.3 returns a JSON, according to the documents linked at issue
        # #592, but it seems that this needs to be enabled(?), otherwise the
        # usual querystring type response is returned.
        try:
            response = response.json()
        except ValueError:
            response = parse_qs(response.text)
        access_token = response["access_token"]
        return self.do_auth(access_token, response, *args, **kwargs)

    def process_refresh_token_response(self, response, *args, **kwargs):
        try:
            return response.json()
        except ValueError:
            return parse_qs(response.content)

    def refresh_token_params(self, token, *args, **kwargs):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "fb_exchange_token": token,
            "grant_type": "fb_exchange_token",
            "client_id": client_id,
            "client_secret": client_secret,
        }

    @handle_http_errors
    def do_auth(self, access_token, response=None, *args, **kwargs):
        response = response or {}

        data = self.user_data(access_token)

        if not isinstance(data, dict):
            # From time to time Facebook responds back a JSON with just
            # False as value, the reason is still unknown, but since the
            # data is needed (it contains the user ID used to identify the
            # account on further logins), this app cannot allow it to
            # continue with the auth process.
            raise AuthUnknownError(
                self, "An error occurred while retrieving users Facebook data"
            )

        data["access_token"] = access_token
        if "expires_in" in response:
            data["expires_in"] = response["expires_in"]

        if self.data.get("granted_scopes"):
            data["granted_scopes"] = self.data["granted_scopes"].split(",")

        if self.data.get("denied_scopes"):
            data["denied_scopes"] = self.data["denied_scopes"].split(",")

        kwargs.update({"backend": self, "response": data})
        return self.strategy.authenticate(*args, **kwargs)

    def revoke_token_url(self, token, uid):
        version = self.setting("API_VERSION", API_VERSION)
        return self.REVOKE_TOKEN_URL.format(version=version, uid=uid)

    def revoke_token_params(self, token, uid):
        return {"access_token": token}

    def process_revoke_token_response(self, response):
        return (
            super().process_revoke_token_response(response)
            and response.content == "true"
        )


class FacebookAppOAuth2(FacebookOAuth2):
    """Facebook Application Authentication support"""

    name = "facebook-app"

    def uses_redirect(self) -> bool:
        return False

    def auth_complete(self, *args, **kwargs):
        access_token = self.data.get("access_token")
        response = {}

        if "signed_request" in self.data:
            _key, _secret = self.get_key_and_secret()
            response = self.load_signed_request(self.data["signed_request"])
            if "user_id" not in response and "oauth_token" not in response:
                raise AuthException(self, "Missing user_id or oauth_token")

            if response is not None:
                access_token = (
                    response.get("access_token")
                    or response.get("oauth_token")
                    or self.data.get("access_token")
                )

        if access_token is None:
            access_error = self.data.get("error")
            if access_error == "access_denied":
                raise AuthCanceled(self)
            raise AuthException(self, access_error)
        return self.do_auth(access_token, response, *args, **kwargs)

    def auth_html(self):
        key, _secret = self.get_key_and_secret()
        namespace = self.setting("NAMESPACE", None)
        scope = self.setting("SCOPE", "")
        if scope:
            scope = self.SCOPE_SEPARATOR.join(scope)
        ctx = {
            "FACEBOOK_APP_NAMESPACE": namespace or key,
            "FACEBOOK_KEY": key,
            "FACEBOOK_EXTENDED_PERMISSIONS": scope,
            "FACEBOOK_COMPLETE_URI": self.redirect_uri,
        }
        tpl = self.setting("LOCAL_HTML", "facebook.html")
        return self.strategy.render_html(tpl=tpl, context=ctx)

    def load_signed_request(self, signed_request) -> dict:
        def base64_url_decode(data):
            data = data.encode("ascii")
            data += b"=" * (4 - (len(data) % 4))
            return base64.urlsafe_b64decode(data)

        _key, secret = self.get_key_and_secret()
        try:
            sig, payload = signed_request.split(".", 1)
        except ValueError as error:
            # ignore if can't split on dot
            raise AuthException(self, "Invalid signed request") from error
        sig = base64_url_decode(sig)
        payload_json_bytes = base64_url_decode(payload)
        data = json.loads(payload_json_bytes.decode("utf-8", "replace"))
        expected_sig = hmac.new(
            secret.encode("ascii"),
            msg=payload.encode("ascii"),
            digestmod=hashlib.sha256,
        ).digest()
        # allow the signed_request to function for upto 1 day
        if constant_time_compare(sig, expected_sig) and data["issued_at"] > (
            time.time() - 86400
        ):
            return data
        raise AuthException(self, "Invalid signature")



from typing import Any
from urllib.parse import urlencode

from .oauth import BaseOAuth2


class SalesforceOAuth2(BaseOAuth2):
    """Salesforce OAuth2 authentication backend"""

    name = "salesforce-oauth2"
    AUTHORIZATION_URL = "https://login.salesforce.com/services/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://login.salesforce.com/services/oauth2/token"
    REVOKE_TOKEN_URL = "https://login.salesforce.com/services/oauth2/revoke"
    REFRESH_TOKEN_METHOD = "POST"
    SCOPE_SEPARATOR = " "
    EXTRA_DATA = [
        ("id", "id"),
        ("instance_url", "instance_url"),
        ("issued_at", "issued_at"),
        ("signature", "signature"),
        ("refresh_token", "refresh_token"),
    ]

    def get_user_details(self, response):
        """Return user details from a Salesforce account"""
        return {
            "username": response.get("username"),
            "email": response.get("email") or "",
            "first_name": response.get("first_name"),
            "last_name": response.get("last_name"),
            "fullname": response.get("display_name"),
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        user_id_url = kwargs["response"]["id"]
        url = f"{user_id_url}?{urlencode({'access_token': access_token})}"
        try:
            return self.get_json(url)
        except ValueError:
            return None


class SalesforceOAuth2Sandbox(SalesforceOAuth2):
    """Salesforce OAuth2 authentication testing backend"""

    name = "salesforce-oauth2-sandbox"
    AUTHORIZATION_URL = "https://test.salesforce.com/services/oauth2/authorize"
    ACCESS_TOKEN_URL = "https://test.salesforce.com/services/oauth2/token"
    REVOKE_TOKEN_URL = "https://test.salesforce.com/services/oauth2/revoke"



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class ThisIsMyJameOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.thisismyjam.ThisIsMyJamOAuth1"
    user_data_url = "http://api.thisismyjam.com/1/verify.json"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps(
        {"id": 10101010, "person": {"name": "foobar", "fullname": "Foo Bar"}}
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Yahoo OpenId, OAuth1 and OAuth2 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/yahoo.html
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from requests.auth import HTTPBasicAuth

from social_core.utils import handle_http_errors

from .oauth import BaseOAuth2

if TYPE_CHECKING:
    from requests.auth import AuthBase


class YahooOAuth2(BaseOAuth2):
    """Yahoo OAuth2 authentication backend"""

    name = "yahoo-oauth2"
    ID_KEY = "sub"
    AUTHORIZATION_URL = "https://api.login.yahoo.com/oauth2/request_auth"
    ACCESS_TOKEN_URL = "https://api.login.yahoo.com/oauth2/get_token"
    EXTRA_DATA = [
        ("sub", "id"),
        ("access_token", "access_token"),
        ("expires_in", "expires_in"),
        ("refresh_token", "refresh_token"),
        ("token_type", "token_type"),
    ]

    def get_user_details(self, response):
        """
        Return user details from Yahoo Profile.
        To Get user email you need the profile private read permission.
        """

        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("given_name"), last_name=response.get("family_name")
        )

        email = response.get("email")
        return {
            "username": response.get("preferred_username", response.get("nickname")),
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""

        url = "https://api.login.yahoo.com/openid/v1/userinfo"

        return self.get_json(
            url, headers={"Authorization": f"Bearer {access_token}"}, method="GET"
        )

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        self.process_error(self.data)
        response = self.request_access_token(
            self.ACCESS_TOKEN_URL,
            auth=self.refresh_token_auth(),
            data=self.auth_complete_params(self.validate_state()),
            headers=self.auth_headers(),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        return self.do_auth(
            response["access_token"], *args, response=response, **kwargs
        )

    def refresh_token_params(self, token, *args, **kwargs):
        return {
            "refresh_token": token,
            "grant_type": "refresh_token",
            "redirect_uri": "oob",  # out of bounds
        }

    def refresh_token_auth(self) -> AuthBase | None:
        return HTTPBasicAuth(*self.get_key_and_secret())

    def auth_complete_params(self, state=None):
        return {
            "grant_type": "authorization_code",  # request auth code
            "code": self.data.get("code", ""),  # server response code
            "redirect_uri": self.get_redirect_uri(state),
        }



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class PinterestOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.pinterest.PinterestOAuth2"
    user_data_url = "https://api.pinterest.com/v1/me/"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "id": "4788400174839062",
            "first_name": "Foo",
            "last_name": "Bar",
            "username": "foobar",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()


class PinterestOAuth2BrokenServerResponseTest(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.pinterest.PinterestOAuth2"
    user_data_url = "https://api.pinterest.com/v1/me/"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "data": {
                "id": "4788400174839062",
                "first_name": "Foo",
                "last_name": "Bar",
                "url": "https://www.pinterest.com/foobar/",
            }
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Backend for Hashicorp Vault OIDC Identity Provider in Vault 1.9+
https://www.vaultproject.io/docs/secrets/identity/oidc-provider
"""

from social_core.backends.open_id_connect import OpenIdConnectAuth


class VaultOpenIdConnect(OpenIdConnectAuth):
    """
    Vault OIDC authentication backend

    This is an alias for the generic OIDC backend
    """

    name = "vault"



"""
Spotify backend, docs at:
    https://developer.spotify.com/spotify-web-api/
    https://developer.spotify.com/spotify-web-api/authorization-guide/
"""

from typing import Any

from .oauth import BaseOAuth2


class SpotifyOAuth2(BaseOAuth2):
    """Spotify OAuth2 authentication backend"""

    name = "spotify"
    ID_KEY = "id"
    AUTHORIZATION_URL = "https://accounts.spotify.com/authorize"
    ACCESS_TOKEN_URL = "https://accounts.spotify.com/api/token"
    SCOPE_SEPARATOR = " "
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("refresh_token", "refresh_token"),
    ]

    def auth_headers(self):
        return {"Authorization": self.get_key_and_secret_basic_auth()}

    def get_user_details(self, response):
        """Return user details from Spotify account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("display_name")
        )
        return {
            "username": response.get("id"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "https://api.spotify.com/v1/me",
            headers={"Authorization": f"Bearer {access_token}"},
        )



"""
Upwork OAuth1 backend
"""

from typing import Any

from .oauth import BaseOAuth1


class UpworkOAuth(BaseOAuth1):
    """Upwork OAuth authentication backend"""

    name = "upwork"
    ID_KEY = "id"
    AUTHORIZATION_URL = "https://www.upwork.com/services/api/auth"
    REQUEST_TOKEN_URL = "https://www.upwork.com/api/auth/v1/oauth/token/request"
    REQUEST_TOKEN_METHOD = "POST"
    ACCESS_TOKEN_URL = "https://www.upwork.com/api/auth/v1/oauth/token/access"
    REDIRECT_URI_PARAMETER_NAME = "oauth_callback"

    def get_user_details(self, response):
        """Return user details from Upwork account"""
        info = response.get("info", {})
        auth_user = response.get("auth_user", {})
        first_name = auth_user.get("first_name")
        last_name = auth_user.get("last_name")
        fullname = f"{first_name} {last_name}"
        profile_url = info.get("profile_url", "")
        username = profile_url.rsplit("/")[-1].replace("~", "")
        return {
            "username": username,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: dict, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "https://www.upwork.com/api/auth/v1/info.json",
            auth=self.oauth_auth(access_token),
        )



"""
SurveyMonkey OAuth2 backend, docs at:
    https://developer.surveymonkey.com/api/v3/#authentication
"""

from typing import Any

from .oauth import BaseOAuth2


class SurveyMonkeyOAuth2(BaseOAuth2):
    """SurveyMonkey OAuth2 authentication backend"""

    name = "surveymonkey"
    AUTHORIZATION_URL = "https://api.surveymonkey.com/oauth/authorize"
    ACCESS_TOKEN_URL = "https://api.surveymonkey.com/oauth/token"
    USER_DATA_URL = "/v3/users/me"
    STATE_PARAMETER = False
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("access_url", "access_url"),
    ]

    def get_user_details(self, response):
        """Return user details from a SurveyMonkey /users/me response"""
        response["name"] = f"{response['first_name']} {response['last_name']}"
        return response

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data information from service"""
        base_url = kwargs["response"]["access_url"]
        return self.get_json(
            base_url + self.USER_DATA_URL,
            headers={"Authorization": f"bearer {access_token}"},
        )



from typing import Any

from requests.auth import HTTPBasicAuth

from social_core.utils import handle_http_errors

from .oauth import BaseOAuth2


class JustGivingOAuth2(BaseOAuth2):
    """Just Giving OAuth authentication backend"""

    name = "justgiving"
    ID_KEY = "userId"
    AUTHORIZATION_URL = "https://identity.justgiving.com/connect/authorize"
    ACCESS_TOKEN_URL = "https://identity.justgiving.com/connect/token"
    USER_DATA_URL = "https://api.justgiving.com/v1/account"
    DEFAULT_SCOPE = ["openid", "account", "profile", "email", "fundraise"]

    def get_user_details(self, response):
        """Return user details from Just Giving account"""
        fullname, first_name, last_name = self.get_user_names(
            "", response.get("firstName"), response.get("lastName")
        )
        return {
            "username": response.get("email"),
            "email": response.get("email"),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        key, secret = self.get_key_and_secret()
        return self.get_json(
            self.USER_DATA_URL,
            headers={
                "Authorization": f"Bearer {access_token}",
                "Content-Type": "application/json",
                "x-application-key": secret,
                "x-api-key": key,
            },
        )

    @handle_http_errors
    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        state = self.validate_state()
        self.process_error(self.data)

        key, secret = self.get_key_and_secret()
        response = self.request_access_token(
            self.access_token_url(),
            data=self.auth_complete_params(state),
            headers=self.auth_headers(),
            auth=HTTPBasicAuth(key, secret),
            method=self.ACCESS_TOKEN_METHOD,
        )
        self.process_error(response)
        return self.do_auth(
            response["access_token"], *args, response=response, **kwargs
        )



from typing import Any

from .goclio import GoClioOAuth2


class GoClioEuOAuth2(GoClioOAuth2):
    name = "goclioeu"
    AUTHORIZATION_URL = "https://app.goclio.eu/oauth/authorize/"
    ACCESS_TOKEN_URL = "https://app.goclio.eu/oauth/token/"

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "https://app.goclio.eu/api/v2/users/who_am_i",
            params={"access_token": access_token},
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from social_core.backends.base import BaseAuth


class SocialAuthBaseException(ValueError):
    """Base class for pipeline exceptions."""


class SocialAuthImproperlyConfiguredError(SocialAuthBaseException):
    """Raised when configuration is invalid."""


class StrategyMissingFeatureError(SocialAuthBaseException):
    """Strategy does not support this."""

    def __init__(self, strategy_name: str, feature_name: str) -> None:
        self.strategy_name = strategy_name
        self.feature_name = feature_name
        super().__init__()

    def __str__(self) -> str:
        return f"Strategy {self.strategy_name} does not support {self.feature_name}"


class DefaultStrategyMissingError(SocialAuthBaseException):
    """Default strategy is not configured."""

    def __str__(self) -> str:
        return "Default strategy is not configured"


class StrategyMissingBackendError(SocialAuthBaseException):
    """Strategy storage backend is not configured."""

    def __str__(self) -> str:
        return "Strategy storage backend is not configured"


class WrongBackend(SocialAuthBaseException):
    def __init__(self, backend_name: str) -> None:
        self.backend_name = backend_name
        super().__init__()

    def __str__(self) -> str:
        return f'Incorrect authentication service "{self.backend_name}"'


class MissingBackend(WrongBackend):
    def __str__(self) -> str:
        return f'Missing backend "{self.backend_name}" entry'


class NotAllowedToDisconnect(SocialAuthBaseException):
    """User is not allowed to disconnect it's social account."""

    def __str__(self) -> str:
        return "This account is not allowed to be disconnected."


class AuthException(SocialAuthBaseException):
    """Auth process exception."""

    def __init__(self, backend: BaseAuth, *args, **kwargs) -> None:
        self.backend = backend
        super().__init__(*args, **kwargs)


class AuthFailed(AuthException):
    """Auth process failed for some reason."""

    def __str__(self) -> str:
        msg = super().__str__()
        if msg == "access_denied":
            return "Authentication process was canceled"
        return f"Authentication failed: {msg}"


class AuthCanceled(AuthException):
    """Auth process was canceled by user."""

    def __init__(self, *args, **kwargs) -> None:
        self.response = kwargs.pop("response", None)
        super().__init__(*args, **kwargs)

    def __str__(self) -> str:
        msg = super().__str__()
        if msg:
            return f"Authentication process canceled: {msg}"
        return "Authentication process canceled"


class AuthUnknownError(AuthException):
    """Unknown auth process error."""

    def __str__(self) -> str:
        msg = super().__str__()
        return f"An unknown error happened while authenticating {msg}"


class AuthTokenError(AuthException):
    """Auth token error."""

    def __str__(self) -> str:
        msg = super().__str__()
        return f"Token error: {msg}"


class AuthMissingParameter(AuthException):
    """Missing parameter needed to start or complete the process."""

    def __init__(self, backend: BaseAuth, parameter: str, *args, **kwargs) -> None:
        self.parameter = parameter
        super().__init__(backend, *args, **kwargs)

    def __str__(self) -> str:
        return f"Missing needed parameter {self.parameter}"


class AuthInvalidParameter(AuthMissingParameter):
    """Invalid value for parameter to start or complete the process."""

    def __str__(self) -> str:
        return f"Invalid value for parameter {self.parameter}"


class AuthNotImplementedParameter(AuthMissingParameter):
    """Optional parameter not implemented to start or complete the process."""

    def __str__(self) -> str:
        return f"Not implemented parameter {self.parameter}"


class AuthStateMissing(AuthException):
    """State parameter is incorrect."""

    def __str__(self) -> str:
        return "Session value state missing."


class AuthStateForbidden(AuthException):
    """State parameter is incorrect."""

    def __str__(self) -> str:
        return "Wrong state parameter given."


class AuthAlreadyAssociated(AuthException):
    """A different user has already associated the target social account"""

    def __str__(self) -> str:
        return "This account is already in use."


class AuthTokenRevoked(AuthException):
    """User revoked the access_token in the provider."""

    def __str__(self) -> str:
        return "User revoke access to the token"


class AuthForbidden(AuthException):
    """Authentication for this user is forbidden"""

    def __str__(self) -> str:
        return "Your credentials aren't allowed"


class AuthUnreachableProvider(AuthException):
    """Cannot reach the provider"""

    def __str__(self) -> str:
        return "The authentication provider could not be reached"


class InvalidEmail(AuthException):
    def __str__(self) -> str:
        return "Email couldn't be validated"


class AuthConnectionError(AuthException):
    """Connection error duing authentication."""

    def __str__(self) -> str:
        msg = super().__str__()
        return f"Connection error: {msg}"


class InvalidExpiryValue(SocialAuthBaseException):
    """Invalid expiry value in extra_data."""

    def __init__(self, field_name: str, value: object) -> None:
        self.field_name = field_name
        self.value = value
        super().__init__()

    def __str__(self) -> str:
        return f"Invalid expiry value for field '{self.field_name}': {self.value}"



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class MineIDOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.mineid.MineIDOAuth2"
    user_data_url = "https://www.mineid.org/api/user"
    expected_username = "foo@bar.com"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "email": "foo@bar.com",
            "primary_profile": None,
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()

    def test_auth_url_parameters(self) -> None:
        self.check_parameters_in_authorization_url("AUTHORIZATION_URL")



from typing import Any, cast

from .oauth import BaseOAuth2


class MineIDOAuth2(BaseOAuth2):
    """MineID OAuth2 authentication backend"""

    name = "mineid"
    AUTHORIZATION_URL = "{scheme}://{host}/oauth/authorize"
    ACCESS_TOKEN_URL = "{scheme}://{host}/oauth/access_token"
    SCOPE_SEPARATOR = ","
    EXTRA_DATA = []

    def get_user_details(self, response):
        """Return user details"""
        return {"email": response.get("email"), "username": response.get("email")}

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self._user_data(access_token)

    def _user_data(self, access_token, path=None):
        url = "{scheme}://{host}/api/user".format(**self.get_mineid_url_params())
        return self.get_json(url, params={"access_token": access_token})

    def get_authorization_url_format(self) -> dict[str, str]:
        return self.get_mineid_url_params()

    def get_access_token_url_format(self) -> dict[str, str]:
        return self.get_mineid_url_params()

    def get_mineid_url_params(self) -> dict[str, str]:
        return {
            "host": cast("str", self.setting("HOST", "www.mineid.org")),
            "scheme": cast("str", self.setting("SCHEME", "https")),
        }



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from social_core.exceptions import InvalidEmail

from .partial import partial

if TYPE_CHECKING:
    from social_core.backends.base import BaseAuth


@partial
def mail_validation(backend: BaseAuth, details, is_new=False, *args, **kwargs):
    requires_validation = backend.REQUIRES_EMAIL_VALIDATION or backend.setting(
        "FORCE_EMAIL_VALIDATION", False
    )
    send_validation = details.get("email") and (
        is_new or backend.setting("PASSWORDLESS", False)
    )
    if requires_validation and send_validation:
        data = backend.strategy.request_data()
        if "verification_code" in data:
            backend.strategy.session_pop("email_validation_address")
            if not backend.strategy.validate_email(
                details["email"], data["verification_code"]
            ):
                raise InvalidEmail(backend)
            return None
        current_partial = kwargs["current_partial"]
        backend.strategy.send_email_validation(
            backend, details["email"], current_partial.token
        )
        backend.strategy.session_set("email_validation_address", details["email"])
        return backend.strategy.redirect(
            cast("str", backend.strategy.setting("EMAIL_VALIDATION_URL"))
        )
    return None



"""
Twilio auth backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/twilio.html
"""

from re import sub
from urllib.parse import urlencode

from .base import BaseAuth


class TwilioAuth(BaseAuth):
    name = "twilio"
    ID_KEY = "AccountSid"

    def get_user_details(self, response):
        """Return twilio details, Twilio only provides AccountSID as
        parameters."""
        # /complete/twilio/?AccountSid=ACc65ea16c9ebd4d4684edf814995b27e
        return {
            "username": response["AccountSid"],
            "email": "",
            "fullname": "",
            "first_name": "",
            "last_name": "",
        }

    def auth_url(self) -> str:
        """Return authorization redirect url."""
        key, _secret = self.get_key_and_secret()
        callback = self.strategy.absolute_uri(self.redirect_uri)
        callback = sub(r"^https", "http", callback)
        query = urlencode({"cb": callback})
        return f"https://www.twilio.com/authorize/{key}?{query}"

    def auth_complete(self, *args, **kwargs):
        """Completes login process, must return user instance"""
        account_sid = self.data.get("AccountSid")
        if not account_sid:
            raise ValueError("No AccountSid returned")
        kwargs.update({"response": self.data, "backend": self})
        return self.strategy.authenticate(*args, **kwargs)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class AngelOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.angel.AngelOAuth2"
    user_data_url = "https://api.angel.co/1/me/"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "facebook_url": "http://www.facebook.com/foobar",
            "bio": None,
            "name": "Foo Bar",
            "roles": [],
            "github_url": None,
            "angellist_url": "https://angel.co/foobar",
            "image": "https://graph.facebook.com/foobar/picture?type=square",
            "linkedin_url": None,
            "locations": [],
            "twitter_url": None,
            "what_ive_built": None,
            "dribbble_url": None,
            "behance_url": None,
            "blog_url": None,
            "aboutme_url": None,
            "follower_count": 0,
            "online_bio_url": None,
            "id": 101010,
        }
    )
    expected_username = "foobar"

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Auth0 implementation based on:
https://auth0.com/docs/quickstart/webapp/django/01-login
"""

import jwt

from .oauth import BaseOAuth2


class Auth0OAuth2(BaseOAuth2):
    """Auth0 OAuth authentication backend"""

    name = "auth0"
    SCOPE_SEPARATOR = " "
    EXTRA_DATA = [("picture", "picture")]

    def api_path(self, path=""):
        """Build API path for Auth0 domain"""
        return f"https://{self.setting('DOMAIN')}/{path}"

    def authorization_url(self):
        return self.api_path("authorize")

    def access_token_url(self):
        return self.api_path("oauth/token")

    def get_user_id(self, details, response):
        """Return current user id."""
        return details["user_id"]

    def get_user_details(self, response):
        # Obtain JWT and the keys to validate the signature
        id_token = response.get("id_token")
        jwks = self.get_json(self.api_path(".well-known/jwks.json"))
        issuer = self.api_path()
        audience = self.setting("KEY")  # CLIENT_ID
        try:
            # it could be a set of JWKs
            keys = jwt.PyJWKSet.from_dict(jwks).keys
        except jwt.PyJWKSetError:
            # let any error raise from here
            # try to get single JWK
            keys = [jwt.PyJWK.from_dict(jwks, "RS256")]

        signature_error = None
        for key in keys:
            try:
                payload = jwt.decode(
                    id_token,
                    key.key,
                    algorithms=["RS256"],
                    audience=audience,
                    issuer=issuer,
                )
            except (jwt.InvalidSignatureError, jwt.InvalidAlgorithmError) as ex:
                signature_error = ex
            else:
                break
        else:
            assert signature_error is not None
            # raise last esception found during iteration
            raise signature_error

        fullname, first_name, last_name = self.get_user_names(payload["name"])
        return {
            "username": payload["nickname"],
            "email": payload["email"],
            "email_verified": payload.get("email_verified", False),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
            "picture": payload["picture"],
            "user_id": payload["sub"],
        }



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class PhabricatorOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.phabricator.PhabricatorOAuth2"
    user_data_url = "https://secure.phabricator.com/api/user.whoami"
    expected_username = "user"
    access_token_body = json.dumps(
        {
            "access_token": "loremipsumdolorsitametenim",
            "token_type": "Bearer",
            "expires_in": 7200,
            "refresh_token": "foobar",
        }
    )

    user_data_body = json.dumps(
        {
            "phid": "PHID-USER-jbfcsj7c6nkt2tv3trb6",
            "userName": "user",
            "realName": "FirstName LastName",
            "image": "https://secure.phabricator.com/file/data/4qjdpmvca4wwkfw2wevc"
            "/PHID-FILE-t37vxezr54fjuvbrblkp/alphanumeric_lato-white_U.png"
            "-_3f674d-255%2C255%2C255%2C0.7.png",
            "uri": "https://secure.phabricator.com/p/user/",
            "roles": ["admin", "verified", "approved", "activated"],
            "primaryEmail": "user@example.com",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()


class PhabricatorCustomDomainOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.phabricator.PhabricatorOAuth2"
    user_data_url = "https://example.com/api/user.whoami"
    expected_username = "user"
    access_token_body = json.dumps(
        {
            "access_token": "loremipsumdolorsitametenim",
            "token_type": "Bearer",
            "expires_in": 7200,
            "refresh_token": "foobar",
        }
    )

    user_data_body = json.dumps(
        {
            "phid": "PHID-USER-jbfcsj7c6nkt2tv3trb6",
            "userName": "user",
            "realName": "FirstName LastName",
            "image": "https://example.com/file/data/4qjdpmvca4wwkfw2wevc/PHID-FILE-"
            "t37vxezr54fjuvbrblkp/alphanumeric_lato-white_U.png-_3f674d-25"
            "5%2C255%2C255%2C0.7.png",
            "uri": "https://example.com/p/user/",
            "roles": ["admin", "verified", "approved", "activated"],
            "primaryEmail": "user@example.com",
        }
    )

    def test_login(self) -> None:
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_PHABRICATOR_API_URL": "https://example.com",
            }
        )
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_PHABRICATOR_API_URL": "https://example.com",
            }
        )
        self.do_partial_pipeline()



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class SkyrockOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.skyrock.SkyrockOAuth"
    user_data_url = "https://api.skyrock.com/v2/user/get.json"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
        }
    )
    user_data_body = json.dumps(
        {
            "locale": "en_US",
            "city": "",
            "has_blog": False,
            "web_messager_enabled": True,
            "email": "foo@bar.com",
            "username": "foobar",
            "firstname": "Foo",
            "user_url": "",
            "address1": "",
            "address2": "",
            "has_profile": False,
            "allow_messages_from": "everybody",
            "is_online": False,
            "postalcode": "",
            "lang": "en",
            "id_user": 10101010,
            "name": "Bar",
            "gender": 0,
            "avatar_url": "http://www.skyrock.com/img/avatars/default-0.jpg",
            "nb_friends": 0,
            "country": "US",
            "birth_date": "1980-06-10",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class YandexOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.yandex.YandexOAuth2"
    user_data_url = "https://login.yandex.ru/info"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "display_name": "foobar",
            "real_name": "Foo Bar",
            "sex": None,
            "id": "101010101",
            "default_email": "foobar@yandex.com",
            "emails": ["foobar@yandex.com"],
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()


class YandexOAuth2TestEmptyEmail(OAuth2Test, BaseAuthUrlTestMixin):
    """
    When user log in to yandex service with social network account (e.g.
    vk.com), they `default_email` could be empty.
    """

    backend_path = "social_core.backends.yandex.YandexOAuth2"
    user_data_url = "https://login.yandex.ru/info"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "display_name": "foobar",
            "real_name": "Foo Bar",
            "sex": None,
            "id": "101010101",
            "default_email": "",
            "emails": [],
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class MusicBrainzAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.musicbrainz.MusicBrainzOAuth2"
    user_data_url = "https://musicbrainz.org/oauth2/userinfo"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "GjtKfJS6G4lupbQcCOiTKo4HcLXUgI1p",
            "expires_in": 3600,
            "token_type": "Bearer",
            "refresh_token": "GjSCBBjp4fnbE0AKo3uFu9qq9K2fFm4u",
        }
    )
    user_data_body = json.dumps(
        {
            "sub": "foobar",
            "email": "foo@bar.com",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class XingOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.xing.XingOAuth"
    user_data_url = "https://api.xing.com/v1/users/me.json"
    expected_username = "FooBar"
    access_token_body = urlencode(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "user_id": "123456_abcdef",
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
        }
    )
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps(
        {
            "users": [
                {
                    "id": "123456_abcdef",
                    "first_name": "Foo",
                    "last_name": "Bar",
                    "display_name": "Foo Bar",
                    "page_name": "Foo_Bar",
                    "permalink": "https://www.xing.com/profile/Foo_Bar",
                    "gender": "m",
                    "birth_date": {"day": 12, "month": 8, "year": 1963},
                    "active_email": "foo@bar.com",
                    "time_zone": {"name": "Europe/Copenhagen", "utc_offset": 2.0},
                    "premium_services": ["SEARCH", "PRIVATEMESSAGES"],
                    "badges": ["PREMIUM", "MODERATOR"],
                    "wants": "Nothing",
                    "haves": "Skills",
                    "interests": "Foo Foo",
                    "organisation_member": "ACM, GI",
                    "languages": {
                        "de": "NATIVE",
                        "en": "FLUENT",
                        "fr": None,
                        "zh": "BASIC",
                    },
                    "private_address": {
                        "city": "Foo",
                        "country": "DE",
                        "zip_code": "20357",
                        "street": "Bar",
                        "phone": "12|34|1234560",
                        "fax": "||",
                        "province": "Foo",
                        "email": "foo@bar.com",
                        "mobile_phone": "12|3456|1234567",
                    },
                    "business_address": {
                        "city": "Foo",
                        "country": "DE",
                        "zip_code": "20357",
                        "street": "Bar",
                        "phone": "12|34|1234569",
                        "fax": "12|34|1234561",
                        "province": "Foo",
                        "email": "foo@bar.com",
                        "mobile_phone": "12|345|12345678",
                    },
                    "web_profiles": {
                        "qype": ["http://qype.de/users/foo"],
                        "google_plus": ["http://plus.google.com/foo"],
                        "blog": ["http://blog.example.org"],
                        "homepage": ["http://example.org", "http://other-example.org"],
                    },
                    "instant_messaging_accounts": {
                        "skype": "foobar",
                        "googletalk": "foobar",
                    },
                    "professional_experience": {
                        "primary_company": {
                            "name": "XING AG",
                            "title": "Softwareentwickler",
                            "company_size": "201-500",
                            "tag": None,
                            "url": "http://www.xing.com",
                            "career_level": "PROFESSIONAL_EXPERIENCED",
                            "begin_date": "2010-01",
                            "description": None,
                            "end_date": None,
                            "industry": "AEROSPACE",
                        },
                        "non_primary_companies": [
                            {
                                "name": "Ninja Ltd.",
                                "title": "DevOps",
                                "company_size": None,
                                "tag": "NINJA",
                                "url": "http://www.ninja-ltd.co.uk",
                                "career_level": None,
                                "begin_date": "2009-04",
                                "description": None,
                                "end_date": "2010-07",
                                "industry": "ALTERNATIVE_MEDICINE",
                            },
                            {
                                "name": None,
                                "title": "Wiss. Mitarbeiter",
                                "company_size": None,
                                "tag": "OFFIS",
                                "url": "http://www.uni.de",
                                "career_level": None,
                                "begin_date": "2007",
                                "description": None,
                                "end_date": "2008",
                                "industry": "APPAREL_AND_FASHION",
                            },
                            {
                                "name": None,
                                "title": "TEST NINJA",
                                "company_size": "201-500",
                                "tag": "TESTCOMPANY",
                                "url": None,
                                "career_level": "ENTRY_LEVEL",
                                "begin_date": "1998-12",
                                "description": None,
                                "end_date": "1999-05",
                                "industry": "ARTS_AND_CRAFTS",
                            },
                        ],
                        "awards": [
                            {
                                "name": "Awesome Dude Of The Year",
                                "date_awarded": 2007,
                                "url": None,
                            }
                        ],
                    },
                    "educational_background": {
                        "schools": [
                            {
                                "name": "Foo University",
                                "degree": "MSc CE/CS",
                                "notes": None,
                                "subject": None,
                                "begin_date": "1998-08",
                                "end_date": "2005-02",
                            }
                        ],
                        "qualifications": ["TOEFLS", "PADI AOWD"],
                    },
                    "photo_urls": {
                        "large": "http://www.xing.com/img/users/e/3/d/"
                        "f94ef165a.123456,1.140x185.jpg",
                        "mini_thumb": "http://www.xing.com/img/users/e/3/d/"
                        "f94ef165a.123456,1.18x24.jpg",
                        "thumb": "http://www.xing.com/img/users/e/3/d/"
                        "f94ef165a.123456,1.30x40.jpg",
                        "medium_thumb": "http://www.xing.com/img/users/e/3/d/"
                        "f94ef165a.123456,1.57x75.jpg",
                        "maxi_thumb": "http://www.xing.com/img/users/e/3/d/"
                        "f94ef165a.123456,1.70x93.jpg",
                    },
                }
            ]
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Reddit OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/reddit.html
"""

from typing import Any, cast

from .oauth import BaseOAuth2


class RedditOAuth2(BaseOAuth2):
    """Reddit OAuth2 authentication backend"""

    name = "reddit"
    AUTHORIZATION_URL = "https://ssl.reddit.com/api/v1/authorize"
    ACCESS_TOKEN_URL = "https://ssl.reddit.com/api/v1/access_token"
    REFRESH_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False
    SCOPE_SEPARATOR = ","
    DEFAULT_SCOPE = ["identity"]
    EXTRA_DATA = [
        ("id", "id"),
        ("name", "username"),
        ("link_karma", "link_karma"),
        ("comment_karma", "comment_karma"),
        ("refresh_token", "refresh_token"),
        ("expires_in", "expires_in"),
    ]

    def get_user_details(self, response):
        """Return user details from Reddit account"""
        return {
            "username": response.get("name"),
            "email": "",
            "fullname": "",
            "first_name": "",
            "last_name": "",
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "https://oauth.reddit.com/api/v1/me.json",
            headers={"Authorization": f"bearer {access_token}"},
        )

    def auth_headers(self):
        return {"Authorization": self.get_key_and_secret_basic_auth()}

    def refresh_token_params(self, token: str, *args, **kwargs) -> dict[str, str]:
        params = super().refresh_token_params(token)
        params["redirect_uri"] = self.redirect_uri or cast(
            "str", kwargs.get("redirect_uri")
        )
        return params

    def auth_complete_credentials(self):
        return self.get_key_and_secret()



import json
from unittest.mock import patch

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class SoundcloudOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.soundcloud.SoundcloudOAuth2"
    user_data_url = "https://api.soundcloud.com/me"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "website": None,
            "myspace_name": None,
            "public_favorites_count": 0,
            "followings_count": 0,
            "full_name": "Foo Bar",
            "id": 10101010,
            "city": None,
            "track_count": 0,
            "playlist_count": 0,
            "discogs_name": None,
            "private_tracks_count": 0,
            "followers_count": 0,
            "online": True,
            "username": "foobar",
            "description": None,
            "subscriptions": [],
            "kind": "user",
            "quota": {
                "unlimited_upload_quota": False,
                "upload_seconds_left": 7200,
                "upload_seconds_used": 0,
            },
            "website_title": None,
            "primary_email_confirmed": False,
            "permalink_url": "http://soundcloud.com/foobar",
            "private_playlists_count": 0,
            "permalink": "foobar",
            "upload_seconds_left": 7200,
            "country": None,
            "uri": "https://api.soundcloud.com/users/10101010",
            "avatar_url": "https://a1.sndcdn.com/images/default_avatar_large.png?ca77017",
            "plan": "Free",
        }
    )

    def test_login(self) -> None:
        """Test standard login flow"""
        assert self.user_data_body is not None
        with patch.object(
            self.backend, "user_data", return_value=json.loads(self.user_data_body)
        ):
            self.do_login()

    def test_partial_pipeline(self) -> None:
        """Test partial pipeline flow"""
        assert self.user_data_body is not None
        with patch.object(
            self.backend, "user_data", return_value=json.loads(self.user_data_body)
        ):
            self.do_partial_pipeline()

    def test_user_data(self) -> None:
        """Test user_data method with Authorization header"""
        self.strategy.set_settings(
            {
                "SOCIAL_AUTH_SOUNDCLOUD_KEY": "test_client_id",
                "SOCIAL_AUTH_SOUNDCLOUD_SECRET": "test_client_secret",
            }
        )

        # Mock the HTTP request to the user data endpoint
        with patch("social_core.backends.base.BaseAuth.request") as mock_request:
            assert self.user_data_body is not None
            mock_request.return_value.json.return_value = json.loads(
                self.user_data_body
            )
            response = self.backend.user_data(access_token="foobar")  # noqa: S106

            # Verify the request was made with the correct parameters
            mock_request.assert_called_once_with(
                "https://api.soundcloud.com/me",
                headers={"Authorization": "OAuth foobar"},
                params={"format": "json"},
                method="GET",
                data=None,
                json=None,
                auth=None,
                timeout=None,
            )

            # Verify the response data
            self.assertEqual(response["username"], self.expected_username)
            self.assertEqual(response["permalink_url"], "http://soundcloud.com/foobar")
            self.assertEqual(response["id"], 10101010)
            self.assertEqual(response["full_name"], "Foo Bar")



import json

from .oauth import BaseAuthUrlTestMixin
from .open_id_connect import OpenIdConnectTest


class FenceOpenIdConnectTest(OpenIdConnectTest, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.fence.Fence"
    issuer = "https://nci-crdc.datacommons.io/"
    openid_config_body = json.dumps(
        {
            "issuer": "https://nci-crdc.datacommons.io/",
            "authorization_endpoint": "https://nci-crdc.datacommons.io/user/oauth2/authorize",
            "userinfo_endpoint": "https://nci-crdc.datacommons.io/user/user/",
            "token_endpoint": "https://nci-crdc.datacommons.io/user/oauth2/token",
            "revocation_endpoint": "https://nci-crdc.datacommons.io/user/oauth2/revoke",
            "jwks_uri": "https://auth.globus.org/jwk.json",
            "response_types_supported": ["code", "token", "token id_token", "id_token"],
            "id_token_signing_alg_values_supported": ["RS512"],
            "scopes_supported": [
                "ga4gh_passport_v1",
                "openid",
                "google_credentials",
                "google_service_account",
                "data",
                "user",
                "google_link",
                "admin",
                "fence",
            ],
            "token_endpoint_auth_methods_supported": ["authorization_code", "implicit"],
            "claims_supported": [
                "aud",
                "sub",
                "iss",
                "exp",
                "jti",
                "auth_time",
                "azp",
                "nonce",
                "context",
            ],
            "subject_types_supported": ["public"],
        }
    )
    skip_invalid_at_hash = allow_invalid_at_hash = True



from google.auth.transport import requests as transport_requests
from google.oauth2 import id_token

from social_core.backends.base import BaseAuth
from social_core.backends.google import BaseGoogleAuth
from social_core.exceptions import AuthException, AuthTokenError


class GoogleOneTap(BaseGoogleAuth, BaseAuth):
    name = "google-onetap"
    CSRF_KEY = "g_csrf_token"
    CREDENTIAL_KEY = "credential"

    def auth_url(self):
        raise AuthException(self, "Cannot start login flow for Google One Tap")

    def verify_csrf(self, request) -> None:
        csrf_token_body = self.data.get(self.CSRF_KEY)
        csrf_token_cookie = request.COOKIES.get(self.CSRF_KEY)

        if not csrf_token_body:
            raise AuthTokenError(self, "Missing csrf token from response")

        # csrf_token_cookie can be missing due to https://issuetracker.google.com/issues/226157137
        if not csrf_token_cookie and self.setting("IGNORE_MISSING_CSRF_COOKIE", False):
            return

        if csrf_token_body != csrf_token_cookie:
            raise AuthTokenError(
                self, "csrf token from cookie and response does not match"
            )

    def get_decoded_info(self):
        try:
            idinfo = id_token.verify_oauth2_token(
                self.data.get(self.CREDENTIAL_KEY),
                transport_requests.Request(),
                self.setting("KEY"),
            )
        except ValueError as error:
            raise AuthException(self, "Invalid response from Google") from error

        return idinfo

    def auth_complete(self, *args, **kwargs):
        self.verify_csrf(kwargs["request"])

        response = self.get_decoded_info()
        kwargs.update({"response": response, "backend": self})

        return self.strategy.authenticate(*args, **kwargs)



"""
ThisIsMyJam OAuth1 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/thisismyjam.html
"""

from typing import Any

from .oauth import BaseOAuth1


class ThisIsMyJamOAuth1(BaseOAuth1):
    """ThisIsMyJam OAuth1 authentication backend"""

    name = "thisismyjam"
    REQUEST_TOKEN_URL = "http://www.thisismyjam.com/oauth/request_token"
    AUTHORIZATION_URL = "http://www.thisismyjam.com/oauth/authorize"
    ACCESS_TOKEN_URL = "http://www.thisismyjam.com/oauth/access_token"
    REDIRECT_URI_PARAMETER_NAME = "oauth_callback"

    def get_user_details(self, response):
        """Return user details from ThisIsMyJam account"""
        info = response.get("person")
        fullname, first_name, last_name = self.get_user_names(info.get("fullname"))
        return {
            "username": info.get("name"),
            "email": "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: dict, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "http://api.thisismyjam.com/1/verify.json",
            auth=self.oauth_auth(access_token),
        )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class MicrosoftOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.microsoft.MicrosoftOAuth2"
    user_data_url = "https://graph.microsoft.com/v1.0/me"
    expected_username = "foobar"
    user_data_body = json.dumps(
        {
            "displayName": "foo bar",
            "givenName": "foobar",
            "jobTitle": "Auditor",
            "mail": "foobar@foobar.com",
            "mobilePhone": None,
            "officeLocation": "12/1110",
            "preferredLanguage": "en-US",
            "surname": "Bowen",
            "userPrincipalName": "foobar",
            "id": "48d31887-5fad-4d73-a9f5-3c356e68a038",
        }
    )
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "id_token": "",
            "expires_in": 3600,
            "expires_on": 1423650396,
            "not_before": 1423646496,
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "expires_in": 3600,
            "refresh_token": "foobar-new-refresh-token",
            "scope": "identity",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        _user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")



import json
from urllib.parse import urlencode

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class StackoverflowOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.stackoverflow.StackoverflowOAuth2"
    user_data_url = "https://api.stackexchange.com/2.1/me"
    expected_username = "foobar"
    access_token_body = urlencode({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "items": [
                {
                    "user_id": 101010,
                    "user_type": "registered",
                    "creation_date": 1278525551,
                    "display_name": "foobar",
                    "profile_image": "http: //www.gravatar.com/avatar/"
                    "5280f15cedf540b544eecc30fcf3027c?"
                    "d=identicon&r=PG",
                    "reputation": 547,
                    "reputation_change_day": 0,
                    "reputation_change_week": 0,
                    "reputation_change_month": 0,
                    "reputation_change_quarter": 65,
                    "reputation_change_year": 65,
                    "age": 22,
                    "last_access_date": 1363544705,
                    "last_modified_date": 1354035327,
                    "is_employee": False,
                    "link": "http: //stackoverflow.com/users/101010/foobar",
                    "location": "Fooland",
                    "account_id": 101010,
                    "badge_counts": {"gold": 0, "silver": 3, "bronze": 6},
                }
            ],
            "quota_remaining": 9997,
            "quota_max": 10000,
            "has_more": False,
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import datetime
import json
from urllib.parse import urlencode

import pytest
import responses

from social_core.exceptions import AuthFailed

from .open_id import OpenIdTest

INFO_URL = "http://api.steampowered.com/ISteamUser/GetPlayerSummaries/v0002/?"
JANRAIN_NONCE = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")


class SteamOpenIdTest(OpenIdTest):
    backend_path = "social_core.backends.steam.SteamOpenId"
    expected_username = "foobar"
    discovery_body = """<?xml version="1.0" encoding="UTF-8"?>
<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">
    <XRD>
        <Service priority="0">
            <Type>http://specs.openid.net/auth/2.0/server</Type>
            <URI>https://steamcommunity.com/openid/login</URI>
        </Service>
    </XRD>
</xrds:XRDS>"""
    user_discovery_body = """<?xml version="1.0" encoding="UTF-8"?>
<xrds:XRDS xmlns:xrds="xri://$xrds" xmlns="xri://$xrd*($v*2.0)">
    <XRD>
        <Service priority="0">
            <Type>http://specs.openid.net/auth/2.0/signon</Type>
            <URI>https://steamcommunity.com/openid/login</URI>
        </Service>
    </XRD>
</xrds:XRDS>"""
    server_response = urlencode(
        {
            "janrain_nonce": JANRAIN_NONCE,
            "openid.ns": "http://specs.openid.net/auth/2.0",
            "openid.mode": "id_res",
            "openid.op_endpoint": "https://steamcommunity.com/openid/login",
            "openid.claimed_id": "https://steamcommunity.com/openid/id/123",
            "openid.identity": "https://steamcommunity.com/openid/id/123",
            "openid.return_to": f"http://myapp.com/complete/steam/?janrain_nonce={JANRAIN_NONCE}",
            "openid.response_nonce": f"{JANRAIN_NONCE}oD4UZ3w9chOAiQXk0AqDipqFYRA=",
            "openid.assoc_handle": "1234567890",
            "openid.signed": "signed,op_endpoint,claimed_id,identity,return_to,"
            "response_nonce,assoc_handle",
            "openid.sig": "1az53vj9SVdiBwhk8%2BFQ68R2plo=",
        }
    )
    player_details = json.dumps(
        {
            "response": {
                "players": [
                    {
                        "steamid": "123",
                        "primaryclanid": "1234",
                        "timecreated": 1360768416,
                        "personaname": "foobar",
                        "personastate": 0,
                        "communityvisibilitystate": 3,
                        "profileurl": "http://steamcommunity.com/profiles/123/",
                        "avatar": "http://media.steampowered.com/steamcommunity/"
                        "public/images/avatars/fe/fef49e7fa7e1997310d7"
                        "05b2a6158ff8dc1cdfeb.jpg",
                        "avatarfull": "http://media.steampowered.com/steamcommunity/"
                        "public/images/avatars/fe/fef49e7fa7e1997310d7"
                        "05b2a6158ff8dc1cdfeb_full.jpg",
                        "avatarmedium": "http://media.steampowered.com/steamcommunity/"
                        "public/images/avatars/fe/fef49e7fa7e1997310d7"
                        "05b2a6158ff8dc1cdfeb_medium.jpg",
                        "lastlogoff": 1360790014,
                    }
                ]
            }
        }
    )

    def _login_setup(self, user_url=None) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_STEAM_API_KEY": "123abc"})
        responses.add(
            responses.POST,
            "https://steamcommunity.com/openid/login",
            status=200,
            body=self.server_response,
        )
        responses.add(
            responses.GET,
            user_url or "https://steamcommunity.com/openid/id/123",
            status=200,
            body=self.user_discovery_body,
        )
        responses.add(responses.GET, INFO_URL, status=200, body=self.player_details)

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_login(self) -> None:
        self._login_setup()
        self.do_login()

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_partial_pipeline(self) -> None:
        self._login_setup()
        self.do_partial_pipeline()


class SteamOpenIdMissingSteamIdTest(SteamOpenIdTest):
    server_response = urlencode(
        {
            "janrain_nonce": JANRAIN_NONCE,
            "openid.ns": "http://specs.openid.net/auth/2.0",
            "openid.mode": "id_res",
            "openid.op_endpoint": "https://steamcommunity.com/openid/login",
            "openid.claimed_id": "https://steamcommunity.com/openid/BROKEN",
            "openid.identity": "https://steamcommunity.com/openid/BROKEN",
            "openid.return_to": f"http://myapp.com/complete/steam/?janrain_nonce={JANRAIN_NONCE}",
            "openid.response_nonce": f"{JANRAIN_NONCE}oD4UZ3w9chOAiQXk0AqDipqFYRA=",
            "openid.assoc_handle": "1234567890",
            "openid.signed": "signed,op_endpoint,claimed_id,identity,return_to,"
            "response_nonce,assoc_handle",
            "openid.sig": "1az53vj9SVdiBwhk8%2BFQ68R2plo=",
        }
    )

    def test_login(self) -> None:
        self._login_setup(user_url="https://steamcommunity.com/openid/BROKEN")
        with self.assertRaises(AuthFailed):
            self.do_login()

    def test_partial_pipeline(self) -> None:
        self._login_setup(user_url="https://steamcommunity.com/openid/BROKEN")
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()


class SteamOpenIdFakeSteamIdTest(SteamOpenIdTest):
    server_response = urlencode(
        {
            "janrain_nonce": JANRAIN_NONCE,
            "openid.ns": "http://specs.openid.net/auth/2.0",
            "openid.mode": "id_res",
            "openid.op_endpoint": "https://steamcommunity.com/openid/login",
            "openid.claimed_id": "https://fakesteamcommunity.com/openid/123",
            "openid.identity": "https://fakesteamcommunity.com/openid/123",
            "openid.return_to": f"http://myapp.com/complete/steam/?janrain_nonce={JANRAIN_NONCE}",
            "openid.response_nonce": f"{JANRAIN_NONCE}oD4UZ3w9chOAiQXk0AqDipqFYRA=",
            "openid.assoc_handle": "1234567890",
            "openid.signed": "signed,op_endpoint,claimed_id,identity,return_to,"
            "response_nonce,assoc_handle",
            "openid.sig": "1az53vj9SVdiBwhk8%2BFQ68R2plo=",
        }
    )

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_login(self) -> None:
        self._login_setup(user_url="https://fakesteamcommunity.com/openid/123")
        with self.assertRaises(AuthFailed):
            self.do_login()

    @pytest.mark.xfail(reason="responses mocking does not work for openid")
    def test_partial_pipeline(self) -> None:
        self._login_setup(user_url="https://fakesteamcommunity.com/openid/123")
        with self.assertRaises(AuthFailed):
            self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class FoursquareOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.foursquare.FoursquareOAuth2"
    user_data_url = "https://api.foursquare.com/v2/users/self"
    expected_username = "FooBar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "notifications": [{"item": {"unreadCount": 0}, "type": "notificationTray"}],
            "meta": {
                "errorType": "deprecated",
                "code": 200,
                "errorDetail": "Please provide an API version to avoid future "
                "errors.See http://bit.ly/vywCav",
            },
            "response": {
                "user": {
                    "photo": "https://is0.4sqi.net/userpix_thumbs/BYKIT01VN4T4BISN.jpg",
                    "pings": False,
                    "homeCity": "Foo, Bar",
                    "id": "1010101",
                    "badges": {"count": 0, "items": []},
                    "friends": {
                        "count": 1,
                        "groups": [
                            {
                                "count": 0,
                                "items": [],
                                "type": "friends",
                                "name": "Mutual friends",
                            },
                            {
                                "count": 1,
                                "items": [
                                    {
                                        "bio": "",
                                        "gender": "male",
                                        "firstName": "Baz",
                                        "relationship": "friend",
                                        "photo": "https://is0.4sqi.net/userpix_thumbs/"
                                        "BYKIT01VN4T4BISN.jpg",
                                        "lists": {
                                            "groups": [
                                                {
                                                    "count": 1,
                                                    "items": [],
                                                    "type": "created",
                                                }
                                            ]
                                        },
                                        "homeCity": "Baz, Qux",
                                        "lastName": "Qux",
                                        "tips": {"count": 0},
                                        "id": "10101010",
                                    }
                                ],
                                "type": "others",
                                "name": "Other friends",
                            },
                        ],
                    },
                    "referralId": "u-1010101",
                    "tips": {"count": 0},
                    "type": "user",
                    "todos": {"count": 0},
                    "bio": "",
                    "relationship": "self",
                    "lists": {"groups": [{"count": 1, "items": [], "type": "created"}]},
                    "photos": {"count": 0, "items": []},
                    "checkinPings": "off",
                    "scores": {"max": 0, "checkinsCount": 0, "goal": 50, "recent": 0},
                    "checkins": {"count": 0},
                    "firstName": "Foo",
                    "gender": "male",
                    "contact": {"email": "foo@bar.com"},
                    "lastName": "Bar",
                    "following": {"count": 0},
                    "requests": {"count": 0},
                    "mayorships": {"count": 0, "items": []},
                }
            },
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin
from .open_id_connect import OpenIdConnectTest


class FedoraOpenIdConnectTest(OpenIdConnectTest, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.fedora.FedoraOpenIdConnect"
    user_data_url = "https://id.fedoraproject.org/openidc/UserInfo"
    issuer = "https://id.fedoraproject.org/openidc/"
    openid_config_body = json.dumps(
        {
            "issuer": "https://id.fedoraproject.org/openidc/",
            "authorization_endpoint": "https://id.fedoraproject.org/openidc/Authorization",
            "token_endpoint": "https://id.fedoraproject.org/openidc/Token",
            "token_introspection_endpoint": "https://id.fedoraproject.org/openidc/TokenInfo",
            "introspection_endpoint": "https://id.fedoraproject.org/openidc/TokenInfo",
            "userinfo_endpoint": "https://id.fedoraproject.org/openidc/UserInfo",
            "jwks_uri": "https://id.fedoraproject.org/openidc/Jwks",
            "scopes_supported": [
                "openid",
                "profile",
                "email",
                "address",
                "phone",
                "https://id.fedoraproject.org/scope/groups",
                "https://id.fedoraproject.org/scope/agreements",
                "https://id.fedoraproject.org/scope/fas-attributes",
                "https://github.com/jmflinuxtx/kerneltest-harness/oidc/upload_test_run",
                "https://src.fedoraproject.org/push",
                "https://waiverdb.fedoraproject.org/oidc/create-waiver",
                "https://fedoraproject.org/wiki/api",
            ],
            "response_types_supported": ["code", "id_token", "token", "token id_token"],
            "response_modes_supported": [
                "query",
                "fragment",
                "form_post",
                "oob",
                "none",
            ],
            "grant_types_supported": [
                "authorization_code",
                "implicit",
                "refresh_token",
            ],
            "acr_values_supported": ["0"],
            "subject_types_supported": ["pairwise", "public"],
            "id_token_signing_alg_values_supported": ["RS256"],
            "id_token_encryption_alg_values_supported": [],
            "id_token_encryption_enc_values_supported": [],
            "userinfo_signing_alg_values_supported": ["RS256"],
            "userinfo_encryption_alg_values_supported": [],
            "userinfo_encryption_enc_values_supported": [],
            "request_object_signing_alg_values_supported": [
                "HS256",
                "HS384",
                "HS512",
                "RS256",
                "RS384",
                "RS512",
                "ES256",
                "ES384",
                "ES512",
                "PS256",
                "PS384",
                "PS512",
                "EdDSA",
                "ES256K",
                "none",
            ],
            "request_object_encryption_alg_values_supported": [],
            "request_object_encryption_enc_values_supported": [],
            "token_endpoint_auth_methods_supported": [
                "client_secret_basic",
                "client_secret_post",
            ],
            "token_endpoint_auth_signing_alg_values_supported": ["RS256"],
            "display_values_supported": ["page", "popup"],
            "claim_types_supported": ["normal"],
            "claims_supported": [
                "sub",
                "name",
                "given_name",
                "family_name",
                "middle_name",
                "nickname",
                "preferred_username",
                "profile",
                "picture",
                "website",
                "email",
                "email_verified",
                "gender",
                "birthdate",
                "zoneinfo",
                "locale",
                "phone_number",
                "phone_number_verified",
                "address",
                "updated_at",
            ],
            "service_documentation": "https://fedoraproject.org/wiki/Infrastructure/Authentication/",
            "ui_locales_supported": ["en"],
            "claims_parameter_supported": True,
            "request_parameter_supported": True,
            "request_uri_parameter_supported": True,
            "require_request_uri_registration": False,
            "op_policy_uri": "https://fedoraproject.org/wiki/Legal:PrivacyPolicy/",
            "op_tos_uri": "https://fedoraproject.org/wiki/Legal:PrivacyPolicy/",
        }
    )



import json
from urllib.parse import urlencode

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class DatagouvfrOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.udata.DatagouvfrOAuth2"
    user_data_url = "https://www.data.gouv.fr/api/1/me/"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "first_name": "foobar",
            "email": "foobar@example.com",
        }
    )
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_body = json.dumps({})

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class ScistarterOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.scistarter.SciStarterOAuth2"
    user_data_url = "https://scistarter.com/api/user_info"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "profile_id": 42006,
            "user_id": 5,
            "url": "https://scistarter.com/user/foobar",
            "result": "success",
            "handle": "foobar",
            "email": "foo@bar.com",
            "first_name": "foo",
            "last_name": "bar",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



from social_core.pipeline.partial import partial


@partial
def ask_for_password(strategy, *args, **kwargs):
    if strategy.session_get("password"):
        return {"password": strategy.session_get("password")}
    return strategy.redirect(strategy.build_absolute_uri("/password"))


@partial
def ask_for_slug(strategy, *args, **kwargs):
    if strategy.session_get("slug"):
        return {"slug": strategy.session_get("slug")}
    return strategy.redirect(strategy.build_absolute_uri("/slug"))


def set_password(strategy, user, *args, **kwargs) -> None:
    user.set_password(kwargs["password"])


def set_slug(strategy, user, *args, **kwargs) -> None:
    user.slug = kwargs["slug"]


def remove_user(strategy, user, *args, **kwargs):
    return {"user": None}


@partial
def set_user_from_kwargs(strategy, *args, **kwargs):
    if strategy.session_get("attribute"):
        assert kwargs["user"].id is not None
        return None
    return strategy.redirect(strategy.build_absolute_uri("/attribute"))


@partial
def set_user_from_args(strategy, user, *args, **kwargs):
    if strategy.session_get("attribute"):
        assert user.id is not None
        return None
    return strategy.redirect(strategy.build_absolute_uri("/attribute"))



import unittest

from social_core.exceptions import (
    DefaultStrategyMissingError,
)
from social_core.registry import REGISTRY

from .strategy import TestStrategy


class StrategyRegistryTestCase(unittest.TestCase):
    def test_missing(self) -> None:
        with self.assertRaises(DefaultStrategyMissingError):
            self.assertIsNotNone(REGISTRY.default_strategy)

    def test_set(self) -> None:
        REGISTRY.default_strategy = TestStrategy(None)
        try:
            self.assertIsInstance(REGISTRY.default_strategy, TestStrategy)
        finally:
            REGISTRY.reset()



from __future__ import annotations

import json

from social_core.exceptions import AuthAlreadyAssociated
from social_core.tests.models import User

from .actions import BaseActionTest


class AssociateActionTest(BaseActionTest):
    expected_username = "foobar"
    user: User

    def setUp(self) -> None:
        super().setUp()
        self.user = User(username="foobar", email="foo@bar.com")
        self.backend.strategy.session_set("username", self.user.username)

    def test_associate(self) -> None:
        self.do_login()
        self.assertEqual(len(self.user.social), 1)
        self.assertEqual(self.user.social[0].provider, "github")

    def test_associate_with_partial_pipeline(self) -> None:
        self.do_login_with_partial_pipeline()
        self.assertEqual(len(self.user.social), 1)
        self.assertEqual(self.user.social[0].provider, "github")


class MultipleAccountsTest(AssociateActionTest):
    alternative_user_data_body = json.dumps(
        {
            "login": "foobar2",
            "id": 2,
            "avatar_url": "https://github.com/images/error/foobar2_happy.gif",
            "gravatar_id": "somehexcode",
            "url": "https://api.github.com/users/foobar2",
            "name": "monalisa foobar2",
            "company": "GitHub",
            "blog": "https://github.com/blog",
            "location": "San Francisco",
            "email": "foo@bar.com",
            "hireable": False,
            "bio": "There once was...",
            "public_repos": 2,
            "public_gists": 1,
            "followers": 20,
            "following": 0,
            "html_url": "https://github.com/foobar2",
            "created_at": "2008-01-14T04:33:35Z",
            "type": "User",
            "total_private_repos": 100,
            "owned_private_repos": 100,
            "private_gists": 81,
            "disk_usage": 10000,
            "collaborators": 8,
            "plan": {
                "name": "Medium",
                "space": 400,
                "collaborators": 10,
                "private_repos": 20,
            },
        }
    )

    def test_multiple_social_accounts(self) -> None:
        self.do_login()
        self.do_login(user_data_body=self.alternative_user_data_body)
        self.assertEqual(len(self.user.social), 2)
        self.assertEqual(self.user.social[0].provider, "github")
        self.assertEqual(self.user.social[1].provider, "github")


class AlreadyAssociatedErrorTest(BaseActionTest):
    def setUp(self) -> None:
        super().setUp()
        self.user1 = User(username="foobar", email="foo@bar.com")
        self.user = None

    def tearDown(self) -> None:
        super().tearDown()
        self.user1 = None
        self.user = None

    def test_already_associated_error(self) -> None:
        self.user = self.user1
        self.do_login()
        self.user = User(username="foobar2", email="foo2@bar2.com")
        with self.assertRaisesRegex(
            AuthAlreadyAssociated, "This account is already in use."
        ):
            self.do_login()



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class DeezerOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.deezer.DeezerOAuth2"
    user_data_url = "http://api.deezer.com/user/me"
    expected_username = "foobar"
    access_token_body = "access_token=foobar&expires=0"
    user_data_body = json.dumps(
        {
            "id": "1",
            "name": "foobar",
            "lastname": "",
            "firstname": "",
            "status": 0,
            "birthday": "1970-01-01",
            "inscription_date": "2015-01-01",
            "gender": "M",
            "link": "https://www.deezer.com/profile/1",
            "picture": "https://api.deezer.com/user/1/image",
            "picture_small": "https://cdns-images.dzcdn.net/images/user//56x56-000000-80-0-0.jpg",
            "picture_medium": "https://cdns-images.dzcdn.net/images/user//250x250-000000-80-0-0.jpg",
            "picture_big": "https://cdns-images.dzcdn.net/images/user//500x500-000000-80-0-0.jpg",
            "country": "FR",
            "lang": "FR",
            "is_kid": False,
            "tracklist": "https://api.deezer.com/user/1/flow",
            "type": "user",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import base64
import unittest
from typing import TYPE_CHECKING, cast
from unittest.mock import Mock, patch

from social_core.backends.base import BaseAuth
from social_core.utils import (
    build_absolute_uri,
    partial_pipeline_data,
    sanitize_redirect,
    slugify,
    user_is_active,
    user_is_authenticated,
)

from .models import TestPartial

if TYPE_CHECKING:
    from social_core.storage import PartialMixin, UserProtocol


class SanitizeRedirectTest(unittest.TestCase):
    def test_none_redirect(self) -> None:
        self.assertEqual(sanitize_redirect(["myapp.com"], None), None)

    def test_empty_redirect(self) -> None:
        self.assertEqual(sanitize_redirect(["myapp.com"], ""), None)

    def test_dict_redirect(self) -> None:
        self.assertEqual(sanitize_redirect(["myapp.com"], {}), None)

    def test_invalid_redirect(self) -> None:
        self.assertEqual(sanitize_redirect(["myapp.com"], {"foo": "bar"}), None)

    def test_wrong_path_redirect(self) -> None:
        self.assertEqual(
            sanitize_redirect(["myapp.com"], "http://notmyapp.com/path/"), None
        )

    def test_invalid_evil_redirect(self) -> None:
        self.assertEqual(sanitize_redirect(["myapp.com"], "///evil.com"), None)

    def test_valid_absolute_redirect(self) -> None:
        self.assertEqual(
            sanitize_redirect(["myapp.com"], "http://myapp.com/path/"),
            "http://myapp.com/path/",
        )

    def test_valid_relative_redirect(self) -> None:
        self.assertEqual(sanitize_redirect(["myapp.com"], "/path/"), "/path/")

    def test_multiple_hosts(self) -> None:
        allowed_hosts = ["myapp1.com", "myapp2.com"]
        for host in allowed_hosts:
            url = f"http://{host}/path/"
            self.assertEqual(sanitize_redirect(allowed_hosts, url), url)

    def test_multiple_hosts_wrong_host(self) -> None:
        self.assertEqual(
            sanitize_redirect(
                ["myapp1.com", "myapp2.com"], "http://notmyapp.com/path/"
            ),
            None,
        )


class UserIsAuthenticatedTest(unittest.TestCase):
    def test_user_is_none(self) -> None:
        self.assertEqual(user_is_authenticated(None), False)

    def test_user_is_not_none(self) -> None:
        self.assertEqual(user_is_authenticated(cast("UserProtocol", object())), True)

    def test_user_has_is_authenticated(self) -> None:
        class User:
            is_authenticated = True

        self.assertEqual(user_is_authenticated(cast("UserProtocol", User())), True)

    def test_user_has_is_authenticated_callable(self) -> None:
        class User:
            def is_authenticated(self) -> bool:
                return True

        self.assertEqual(user_is_authenticated(cast("UserProtocol", User())), True)


class UserIsActiveTest(unittest.TestCase):
    def test_user_is_none(self) -> None:
        self.assertEqual(user_is_active(None), False)

    def test_user_is_not_none(self) -> None:
        self.assertEqual(user_is_active(cast("UserProtocol", object())), True)

    def test_user_has_is_active(self) -> None:
        class User:
            is_active = True

        self.assertEqual(user_is_active(cast("UserProtocol", User())), True)

    def test_user_has_is_active_callable(self) -> None:
        class User:
            def is_active(self) -> bool:
                return True

        self.assertEqual(user_is_active(cast("UserProtocol", User())), True)


class SlugifyTest(unittest.TestCase):
    def test_slugify_formats(self) -> None:
        self.assertEqual(slugify("FooBar"), "foobar")
        self.assertEqual(slugify("Foo Bar"), "foo-bar")
        self.assertEqual(slugify("Foo (Bar)"), "foo-bar")


class BuildAbsoluteURITest(unittest.TestCase):
    host = "http://foobar.com"

    def test_path_none(self) -> None:
        self.assertEqual(build_absolute_uri(self.host), self.host)

    def test_path_empty(self) -> None:
        self.assertEqual(build_absolute_uri(self.host, ""), self.host)

    def test_path_http(self) -> None:
        self.assertEqual(
            build_absolute_uri(self.host, "http://barfoo.com"), "http://barfoo.com"
        )

    def test_path_https(self) -> None:
        self.assertEqual(
            build_absolute_uri(self.host, "https://barfoo.com"), "https://barfoo.com"
        )

    def test_host_ends_with_slash_and_path_starts_with_slash(self) -> None:
        assert self.host, "Subclasses must set the host attribute"
        self.assertEqual(
            build_absolute_uri(f"{self.host}/", "/foo/bar"), "http://foobar.com/foo/bar"
        )

    def test_absolute_uri(self) -> None:
        self.assertEqual(
            build_absolute_uri(self.host, "/foo/bar"), "http://foobar.com/foo/bar"
        )


class PartialPipelineData(unittest.TestCase):
    def test_returns_partial_when_uid_and_email_do_match(self) -> None:
        email = "foo@example.com"
        backend = self._backend({"uid": email})
        backend.strategy.request_data.return_value = {backend.ID_KEY: email}
        key, val = ("foo", "bar")
        partial = cast(
            "PartialMixin", partial_pipeline_data(backend, None, *(), **{key: val})
        )
        self.assertIsNotNone(partial)
        self.assertIn(key, partial.kwargs)
        self.assertEqual(partial.kwargs[key], val)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def test_clean_pipeline_when_uid_does_not_match(self) -> None:
        backend = self._backend({"uid": "foo@example.com"})
        backend.strategy.request_data.return_value = {backend.ID_KEY: "bar@example.com"}
        key, val = ("foo", "bar")
        partial = partial_pipeline_data(backend, None, *(), **{key: val})
        self.assertIsNone(partial)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 1)

    def test_kwargs_included_in_result(self) -> None:
        backend = self._backend()
        key, val = ("foo", "bar")
        partial = cast(
            "PartialMixin", partial_pipeline_data(backend, None, *(), **{key: val})
        )
        self.assertIsNotNone(partial)
        self.assertIn(key, partial.kwargs)
        self.assertEqual(partial.kwargs[key], val)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def test_update_user(self) -> None:
        user = cast("UserProtocol", object())
        backend = self._backend(session_kwargs={"user": None})
        partial = cast("PartialMixin", partial_pipeline_data(backend, user))
        self.assertIsNotNone(partial)
        self.assertIn("user", partial.kwargs)
        self.assertEqual(partial.kwargs["user"], user)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def test_configurable_id_key(self) -> None:
        """Test that ID_KEY can be configured via settings"""
        email = "foo@example.com"
        backend = self._backend({"uid": email})
        # Configure a different ID_KEY via id_key() method
        backend.id_key.return_value = "custom_id"
        backend.strategy.request_data.return_value = {"custom_id": email}
        key, val = ("foo", "bar")
        partial = cast(
            "PartialMixin", partial_pipeline_data(backend, None, *(), **{key: val})
        )
        self.assertIsNotNone(partial)
        self.assertIn(key, partial.kwargs)
        self.assertEqual(partial.kwargs[key], val)
        self.assertEqual(backend.strategy.clean_partial_pipeline.call_count, 0)

    def _backend(self, session_kwargs=None):
        backend = Mock()
        backend.ID_KEY = "email"
        backend.name = "mock-backend"
        backend.id_key.return_value = "email"

        strategy = Mock()
        strategy.request = None
        strategy.request_data.return_value = {}
        strategy.session_get.return_value = object()
        strategy.partial_load.return_value = TestPartial.prepare(
            backend.name, 0, {"args": [], "kwargs": session_kwargs or {}}
        )

        backend.strategy = strategy
        return backend


class GetKeyAndSecretBasicAuthTest(unittest.TestCase):
    def test_basic_auth_returns_bytes(self) -> None:
        """Test that method returns bytes with base64 encoding"""
        test_setting = {"KEY": "test_key", "SECRET": "test_secret"}
        backend = BaseAuth(strategy=Mock())
        with patch("social_core.backends.base.BaseAuth.setting", new=test_setting.get):
            result = backend.get_key_and_secret_basic_auth()
            expected = b"Basic " + base64.b64encode(b"test_key:test_secret")
            self.assertEqual(result, expected)
            self.assertIsInstance(result, bytes)


class IdKeyConfigurabilityTest(unittest.TestCase):
    """Test that ID_KEY is configurable via settings"""

    def test_id_key_uses_class_attribute_by_default(self) -> None:
        """Test that id_key() returns class attribute when no setting is provided"""
        strategy = Mock()
        strategy.setting = Mock(return_value=None)
        backend = BaseAuth(strategy=strategy)
        backend.ID_KEY = "default_id"

        result = backend.id_key()

        self.assertEqual(result, "default_id")
        strategy.setting.assert_called_once_with(
            "ID_KEY", default=None, backend=backend
        )

    def test_id_key_uses_setting_when_provided(self) -> None:
        """Test that id_key() returns setting value when provided"""
        strategy = Mock()
        strategy.setting = Mock(return_value="custom_id")
        backend = BaseAuth(strategy=strategy)
        backend.ID_KEY = "default_id"

        result = backend.id_key()

        self.assertEqual(result, "custom_id")
        strategy.setting.assert_called_once_with(
            "ID_KEY", default=None, backend=backend
        )

    def test_get_user_id_uses_configurable_id_key(self) -> None:
        """Test that get_user_id() uses the configurable id_key()"""
        strategy = Mock()
        strategy.setting = Mock(return_value="custom_user_id")
        backend = BaseAuth(strategy=strategy)
        backend.ID_KEY = "default_id"

        response = {"custom_user_id": "12345", "default_id": "67890"}
        result = backend.get_user_id({}, response)

        self.assertEqual(result, "12345")
        strategy.setting.assert_called_with("ID_KEY", default=None, backend=backend)



"""
VK.com OpenAPI, OAuth2 and Iframe application OAuth2 backends, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/vk.html
"""

from __future__ import annotations

import json
from hashlib import md5
from time import time
from typing import Any, cast

from social_core.exceptions import AuthException, AuthTokenRevoked
from social_core.utils import parse_qs

from .base import BaseAuth
from .oauth import BaseOAuth2


def vk_sig(payload: str) -> str:
    """
    Calculates signature using md5.

    https://dev.vk.com/en/api/open-api/getting-started#Authorization%20on%20the%20Remote%20Side
    """
    return md5(payload.encode("utf-8")).hexdigest()  # noqa: S324


class VKontakteOpenAPI(BaseAuth):
    """VK.COM OpenAPI authentication backend"""

    name = "vk-openapi"
    ID_KEY = "id"

    def get_user_details(self, response):
        """Return user details from VK.com request"""
        nickname = response.get("nickname") or ""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("first_name", [""])[0],
            last_name=response.get("last_name", [""])[0],
        )
        return {
            "username": response["id"] if len(nickname) == 0 else nickname,
            "email": "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        return self.data

    def auth_html(self) -> str:
        """Returns local VK authentication page, not necessary for
        VK to authenticate.
        """
        ctx = {
            "VK_APP_ID": self.setting("APP_ID"),
            "VK_COMPLETE_URL": self.redirect_uri,
        }
        local_html = self.setting("LOCAL_HTML", "vkontakte.html")
        return self.strategy.render_html(tpl=local_html, context=ctx)

    def auth_complete(self, *args, **kwargs):
        """Performs check of authentication in VKontakte, returns User if
        succeeded"""
        session_value = self.strategy.session_get(
            f"vk_app_{cast('str', self.setting('APP_ID'))}"
        )
        if "id" not in self.data or not session_value:
            raise ValueError("VK.com authentication is not completed")

        mapping = parse_qs(session_value)
        check_str = "".join(
            f"{item}={mapping[item]}" for item in ["expire", "mid", "secret", "sid"]
        )

        _key, secret = self.get_key_and_secret()
        vk_hash = vk_sig(check_str + secret)
        if vk_hash != mapping["sig"] or int(mapping["expire"]) < time():
            raise ValueError("VK.com authentication failed: Invalid Hash")

        kwargs.update({"backend": self, "response": self.user_data(mapping["mid"])})
        return self.strategy.authenticate(*args, **kwargs)

    def uses_redirect(self) -> bool:
        """VK.com does not require visiting server url in order
        to do authentication, so auth_xxx methods are not needed to be called.
        Their current implementation is just an example"""
        return False


class VKOAuth2(BaseOAuth2):
    """VKOAuth2 authentication backend"""

    name = "vk-oauth2"
    ID_KEY = "id"
    AUTHORIZATION_URL = "https://oauth.vk.ru/authorize"
    ACCESS_TOKEN_URL = "https://oauth.vk.ru/access_token"
    EXTRA_DATA = [("id", "id"), ("expires_in", "expires_in")]

    def get_user_details(self, response):
        """Return user details from VK.com account"""
        fullname, first_name, last_name = self.get_user_names(
            first_name=response.get("first_name"), last_name=response.get("last_name")
        )
        return {
            "username": response.get("screen_name"),
            "email": response.get("email", ""),
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        request_data = [
            "first_name",
            "last_name",
            "screen_name",
            "nickname",
            "photo",
            *cast("list[str]", self.setting("EXTRA_DATA", [])),
        ]

        fields = ",".join(set(request_data))
        response = self.vk_api(
            "users.get",
            {
                "access_token": access_token,
                "fields": fields,
            },
        )

        if response and response.get("error"):
            error = response["error"]
            msg = error.get("error_msg", "Unknown error")
            if error.get("error_code") == 5:
                raise AuthTokenRevoked(self, msg)
            raise AuthException(self, msg)

        if response:
            data = cast("list[dict[str, str | None]]", response.get("response"))[0]
            data["user_photo"] = data.get("photo")  # Backward compatibility
            return data
        return {}

    def vk_api(self, method: str, data: dict[str, str]) -> dict[Any, Any] | None:
        """
        Calls VK.com OpenAPI method, check:
            https://vk.com/apiclub
            http://goo.gl/yLcaa
        """
        # We need to perform server-side call if no access_token
        data["v"] = cast("str", self.setting("API_VERSION", "5.131"))
        if "access_token" not in data:
            key, secret = self.get_key_and_secret()
            if "api_id" not in data:
                data["api_id"] = key

            data["method"] = method
            data["format"] = "json"
            url = "https://api.vk.ru/api.php"
            param_list = sorted(f"{item}={data[item]}" for item in data)
            data["sig"] = vk_sig("".join(param_list) + secret)
        else:
            url = f"https://api.vk.ru/method/{method}"

        try:
            return self.get_json(url, params=data)
        except (TypeError, KeyError, OSError, ValueError, IndexError):
            return None


class VKAppOAuth2(VKOAuth2):
    """VK.com Application Authentication support"""

    name = "vk-app"

    def auth_complete(self, *args, **kwargs):
        required_params = ("is_app_user", "viewer_id", "access_token", "api_id")
        if not all(param in self.data for param in required_params):
            return None

        auth_key = self.data.get("auth_key")

        # Verify signature, if present
        key, secret = self.get_key_and_secret()
        if auth_key:
            check_key = vk_sig(f"{key}_{self.data.get('viewer_id')}_{secret}")
            if check_key != auth_key:
                raise ValueError("VK.com authentication failed: invalid auth key")

        user_check = self.setting("USERMODE")
        user_id = self.data.get("viewer_id")
        if user_check is not None:
            user_check = int(user_check)
            is_user = 0
            if user_check == 1:
                is_user = self.data.get("is_app_user", 0)
            elif user_check == 2:
                response = self.vk_api("isAppUser", {"user_id": user_id})
                if response is None:
                    return None
                is_user = response.get("response", 0)
            if not int(is_user):
                return None

        request = self.strategy.request_data()
        response = {self.id_key(): user_id}
        response.update(json.loads(request["api_result"])["response"][0])
        return self.strategy.authenticate(
            auth=self,
            backend=self,
            request=request,
            response=response,
        )



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class FitbitOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.fitbit.FitbitOAuth1"
    expected_username = "foobar"
    access_token_body = urlencode(
        {
            "oauth_token_secret": "a-secret",
            "encoded_user_id": "101010",
            "oauth_token": "foobar",
        }
    )
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_url = "https://api.fitbit.com/1/user/-/profile.json"
    user_data_body = json.dumps(
        {
            "user": {
                "weightUnit": "en_US",
                "strideLengthWalking": 0,
                "displayName": "foobar",
                "weight": 62.6,
                "foodsLocale": "en_US",
                "heightUnit": "en_US",
                "locale": "en_US",
                "gender": "NA",
                "memberSince": "2011-12-26",
                "offsetFromUTCMillis": -25200000,
                "height": 0,
                "timezone": "America/Los_Angeles",
                "dateOfBirth": "",
                "encodedId": "101010",
                "avatar": "http://www.fitbit.com/images/profile/"
                "defaultProfile_100_male.gif",
                "waterUnit": "en_US",
                "distanceUnit": "en_US",
                "glucoseUnit": "en_US",
                "strideLengthRunning": 0,
            }
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Stackoverflow OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/stackoverflow.html
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

from social_core.utils import parse_qs, wrap_access_token_error

from .oauth import BaseOAuth2

if TYPE_CHECKING:
    from collections.abc import Mapping

    from requests.auth import AuthBase


class StackoverflowOAuth2(BaseOAuth2):
    """Stackoverflow OAuth2 authentication backend"""

    name = "stackoverflow"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://stackexchange.com/oauth"
    ACCESS_TOKEN_URL = "https://stackexchange.com/oauth/access_token"
    SCOPE_SEPARATOR = ","
    EXTRA_DATA = [("id", "id"), ("expires", "expires")]

    def get_user_details(self, response):
        """Return user details from Stackoverflow account"""
        fullname, first_name, last_name = self.get_user_names(
            response.get("display_name")
        )
        return {
            "username": response.get("link").rsplit("/", 1)[-1],
            "full_name": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "https://api.stackexchange.com/2.1/me",
            params={
                "site": "stackoverflow",
                "access_token": access_token,
                "key": self.setting("API_KEY"),
            },
        )["items"][0]

    def request_access_token(
        self,
        url: str,
        method: Literal["GET", "POST", "DELETE"] = "GET",
        headers: Mapping[str, str | bytes] | None = None,
        data: dict | None = None,
        json: dict | None = None,
        auth: tuple[str, str] | AuthBase | None = None,
        params: dict | None = None,
    ) -> dict[Any, Any]:
        with wrap_access_token_error(self):
            response = self.request(
                url,
                method=method,
                headers=headers,
                data=data,
                json=json,
                auth=auth,
                params=params,
            )
        return parse_qs(response.content)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class UniverseAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.universe.UniverseOAuth2"
    user_data_url = "https://www.universe.com/api/v2/current_user"
    expected_username = "scott+awesome@universe.com"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    user_data_body = json.dumps(
        {
            "current_user": {
                "id": "123456",
                "slug": "foo-bar",
                "first_name": "Scott",
                "last_name": "Vitale",
                "created_at": "2019-01-08T15:49:42.514Z",
                "updated_at": "2019-01-17T19:41:39.711Z",
                "email": "scott+awesome@universe.com",
            }
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import unittest

from social_core.backends.base import BaseAuth
from social_core.tests.models import TestStorage
from social_core.tests.strategy import TestStrategy


class BrokenBackendAuth(BaseAuth):
    name = "broken"


class BrokenBackendTest(unittest.TestCase):
    backend: BrokenBackendAuth

    def setUp(self) -> None:
        self.backend = BrokenBackendAuth(TestStrategy(TestStorage))

    def tearDown(self) -> None:
        del self.backend

    def test_auth_url(self) -> None:
        with self.assertRaisesRegex(NotImplementedError, "Implement in subclass"):
            self.backend.auth_url()

    def test_auth_html(self) -> None:
        self.assertEqual(self.backend.auth_html(), "Implement in subclass")

    def test_auth_complete(self) -> None:
        with self.assertRaisesRegex(NotImplementedError, "Implement in subclass"):
            self.backend.auth_complete()

    def test_get_user_details(self) -> None:
        with self.assertRaisesRegex(NotImplementedError, "Implement in subclass"):
            self.backend.get_user_details(None)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class NaverOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.naver.NaverOAuth2"
    user_data_url = "https://openapi.naver.com/v1/nid/me"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
        }
    )

    user_data_content_type = "text/json"
    user_data_body = json.dumps(
        {
            "resultcode": "00",
            "message": "success",
            "response": {
                "email": "openapi@naver.com",
                "nickname": "foobar",
                "profile_image": "https://ssl.pstatic.net/static/pwe/address/nodata_33x33.gif",
                "age": "40-49",
                "gender": "F",
                "id": "32742776",
                "name": "foobar",
                "birthday": "10-01",
            },
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



"""
Evernote OAuth1 backend (with sandbox mode support), docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/evernote.html
"""

from __future__ import annotations

from typing import Any

from requests import HTTPError

from social_core.exceptions import AuthCanceled

from .oauth import BaseOAuth1


class EvernoteOAuth(BaseOAuth1):
    """
    Evernote OAuth authentication backend.

    Possible Values:
       {'edam_expires': ['1367525289541'],
        'edam_noteStoreUrl': [
            'https://sandbox.evernote.com/shard/s1/notestore'
        ],
        'edam_shard': ['s1'],
        'edam_userId': ['123841'],
        'edam_webApiUrlPrefix': ['https://sandbox.evernote.com/shard/s1/'],
        'oauth_token': [
            'S=s1:U=1e3c1:E=13e66dbee45:C=1370f2ac245:P=185:A=my_user:' \
            'H=411443c5e8b20f8718ed382a19d4ae38'
        ]}
    """

    name = "evernote"
    ID_KEY = "edam_userId"
    AUTHORIZATION_URL = "https://www.evernote.com/OAuth.action"
    REQUEST_TOKEN_URL = "https://www.evernote.com/oauth"
    ACCESS_TOKEN_URL = "https://www.evernote.com/oauth"
    ACCESS_TOKEN_METHOD = "GET"
    EXTRA_DATA = [
        ("access_token", "access_token"),
        ("oauth_token", "oauth_token"),
        ("edam_noteStoreUrl", "store_url"),
        ("edam_expires", "expires"),
    ]

    def get_user_details(self, response):
        """Return user details from Evernote account"""
        return {"username": response["edam_userId"], "email": ""}

    def access_token(self, token):
        """Return request for access token value"""
        try:
            return self.get_querystring(
                self.ACCESS_TOKEN_URL, auth=self.oauth_auth(token)
            )
        except HTTPError as err:
            # Evernote returns a 401 error when AuthCanceled
            if err.response.status_code == 401:
                raise AuthCanceled(self, response=err.response) from err
            raise

    def extra_data(
        self,
        user,
        uid: str,
        response: dict[str, Any],
        details: dict[str, Any],
        pipeline_kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        data = super().extra_data(user, uid, response, details, pipeline_kwargs)
        # Evernote returns expiration timestamp in milliseconds, so it needs to
        # be normalized.
        if "expires" in data:
            data["expires"] = int(data["expires"]) / 1000
        return data

    def user_data(self, access_token: dict, *args, **kwargs) -> dict[str, Any] | None:
        """Return user data provided"""
        return access_token.copy()


class EvernoteSandboxOAuth(EvernoteOAuth):
    name = "evernote-sandbox"
    AUTHORIZATION_URL = "https://sandbox.evernote.com/OAuth.action"
    REQUEST_TOKEN_URL = "https://sandbox.evernote.com/oauth"
    ACCESS_TOKEN_URL = "https://sandbox.evernote.com/oauth"



from typing import Any

from .oauth import BaseOAuth2


class PayPalOAuth2(BaseOAuth2):
    """
    PayPal OAuth2 backend, docs at:
        https://developer.paypal.com/docs/connect-with-paypal/integrate/
    """

    name = "paypal-oauth2"
    ID_KEY = "user_id"
    AUTHORIZATION_URL = "https://www.paypal.com/connect"
    ACCESS_TOKEN_URL = "https://api.paypal.com/v1/oauth2/token"
    USER_DATA_URL = (
        "https://api.paypal.com/v1/identity/oauth2/userinfo?schema=paypalv1.1"
    )
    DEFAULT_SCOPE = ["openid", "profile"]
    REFRESH_TOKEN_METHOD = "POST"
    REDIRECT_STATE = False

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        auth_header = {"Authorization": f"Bearer {access_token}"}
        return self.get_json(self.USER_DATA_URL, headers=auth_header)

    def get_user_details(self, response):
        username = response.get(self.id_key()).split("/")[-1]
        fullname, first_name, last_name = self.get_user_names(
            response.get("name", ""),
            response.get("given_name", ""),
            response.get("family_name", ""),
        )
        emails = response.get("emails", [])
        email = self.get_email(emails)
        return {
            "username": username,
            "email": email,
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def auth_complete_params(self, state=None):
        return {
            "grant_type": "authorization_code",
            "code": self.data.get("code", ""),
        }

    def auth_headers(self):
        return {"Authorization": self.get_key_and_secret_basic_auth()}

    def refresh_token_params(self, token, *args, **kwargs):
        return {"refresh_token": token, "grant_type": "refresh_token"}

    @staticmethod
    def get_email(emails):
        if not emails:
            return ""
        primary_emails = (email for email in emails if email.get("primary", False))
        primary_or_first = next(primary_emails, emails[0])
        return primary_or_first.get("value")


class PayPalOAuth2Sandbox(PayPalOAuth2):
    name = "paypal-oauth2-sandbox"
    AUTHORIZATION_URL = "https://www.sandbox.paypal.com/connect"
    ACCESS_TOKEN_URL = "https://api.sandbox.paypal.com/v1/oauth2/token"
    USER_DATA_URL = (
        "https://api.sandbox.paypal.com/v1/identity/oauth2/userinfo?schema=paypalv1.1"
    )



import json
from unittest.mock import patch

from .oauth import BaseAuthUrlTestMixin, OAuth2Test

TEST_KEY = """
-----BEGIN EC PRIVATE KEY-----
MHcCAQEEIKQya8aIoeoOLeThk7Ad/lLyAo2fTp9IuhIpy2CivH/qoAoGCCqGSM49
AwEHoUQDQgAEyEY7IMlNJtyaF/pdcM/PpQ8OCe19Sf1Yxq4HQsrB2b7QogB95Vjt
6mTZDAhlXIBtuM/JLrdkMfPmwjVKLgxHAQ==
-----END EC PRIVATE KEY-----
"""


token_data = {
    "sub": "11011110101011011011111011101111",
    "first_name": "Foo",
    "last_name": "Bar",
    "email": "foobar@apple.com",
}


class AppleIdTest(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.apple.AppleIdAuth"
    user_data_url = "https://appleid.apple.com/auth/authorize/"
    id_token = "a-id-token"
    access_token_body = json.dumps(
        {"id_token": id_token, "access_token": "a-test-token"}
    )
    expected_username = token_data["sub"]

    def extra_settings(self):
        assert self.name, "Name must be set in subclasses"
        return {
            f"SOCIAL_AUTH_{self.name}_TEAM": "a-team-id",
            f"SOCIAL_AUTH_{self.name}_KEY": "a-key-id",
            f"SOCIAL_AUTH_{self.name}_CLIENT": "a-client-id",
            f"SOCIAL_AUTH_{self.name}_SECRET": TEST_KEY,
            f"SOCIAL_AUTH_{self.name}_SCOPE": ["name", "email"],
        }

    def test_login(self) -> None:
        with patch(
            f"{self.backend_path}.decode_id_token",
            return_value=token_data,
        ) as decode_mock:
            self.do_login()
        assert decode_mock.called
        assert decode_mock.call_args[0] == (self.id_token,)

    def test_partial_pipeline(self) -> None:
        with patch(
            f"{self.backend_path}.decode_id_token",
            return_value=token_data,
        ) as decode_mock:
            self.do_partial_pipeline()
        assert decode_mock.called
        assert decode_mock.call_args[0] == (self.id_token,)



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class YammerOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.yammer.YammerOAuth2"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": {
                "user_id": 1010101010,
                "view_groups": True,
                "modify_messages": True,
                "network_id": 101010,
                "created_at": "2013/03/17 16:39:56 +0000",
                "view_members": True,
                "authorized_at": "2013/03/17 16:39:56 +0000",
                "view_subscriptions": True,
                "view_messages": True,
                "modify_subscriptions": True,
                "token": "foobar",
                "expires_at": None,
                "network_permalink": "foobar.com",
                "view_tags": True,
                "network_name": "foobar.com",
            },
            "user": {
                "last_name": "Bar",
                "web_url": "https://www.yammer.com/foobar/users/foobar",
                "expertise": None,
                "full_name": "Foo Bar",
                "timezone": "Pacific Time (US & Canada)",
                "mugshot_url": "https://mug0.assets-yammer.com/mugshot/images/"
                "48x48/no_photo.png",
                "guid": None,
                "network_name": "foobar",
                "id": 1010101010,
                "previous_companies": [],
                "first_name": "Foo",
                "stats": {"following": 0, "followers": 0, "updates": 1},
                "hire_date": None,
                "state": "active",
                "location": None,
                "department": "Software Development",
                "type": "user",
                "show_ask_for_photo": True,
                "job_title": "Software Developer",
                "interests": None,
                "kids_names": None,
                "activated_at": "2013/03/17 16:27:50 +0000",
                "verified_admin": "false",
                "can_broadcast": "false",
                "schools": [],
                "admin": "false",
                "network_domains": ["foobar.com"],
                "name": "foobar",
                "external_urls": [],
                "url": "https://www.yammer.com/api/v1/users/1010101010",
                "settings": {"xdr_proxy": "https://xdrproxy.yammer.com"},
                "summary": None,
                "network_id": 101010,
                "contact": {
                    "phone_numbers": [],
                    "im": {"username": "", "provider": ""},
                    "email_addresses": [{"type": "primary", "address": "foo@bar.com"}],
                    "has_fake_email": False,
                },
                "birth_date": "",
                "mugshot_url_template": "https://mug0.assets-yammer.com/mugshot/"
                "images/{width}x{height}/no_photo.png",
                "significant_other": None,
            },
            "network": {
                "show_upgrade_banner": False,
                "header_text_color": "#FFFFFF",
                "is_org_chart_enabled": True,
                "name": "foobar.com",
                "is_group_enabled": True,
                "header_background_color": "#396B9A",
                "created_at": "2012/12/26 16:52:35 +0000",
                "profile_fields_config": {
                    "enable_work_phone": True,
                    "enable_mobile_phone": True,
                    "enable_job_title": True,
                },
                "permalink": "foobar.com",
                "paid": False,
                "id": 101010,
                "is_chat_enabled": True,
                "web_url": "https://www.yammer.com/foobar.com",
                "moderated": False,
                "community": False,
                "type": "network",
                "navigation_background_color": "#38699F",
                "navigation_text_color": "#FFFFFF",
            },
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json

import responses

from social_core.tests.backends.oauth import BaseAuthUrlTestMixin
from social_core.tests.backends.open_id_connect import OpenIdConnectTest


class Auth0OpenIdConnectTest(OpenIdConnectTest, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.auth0_openidconnect.Auth0OpenIdConnectAuth"
    domain = "example.auth0.com"
    issuer = f"https://{domain}/"

    openid_config_body = json.dumps(
        {
            "issuer": issuer,
            "authorization_endpoint": f"https://{domain}/authorize",
            "token_endpoint": f"https://{domain}/oauth/token",
            "userinfo_endpoint": f"https://{domain}/userinfo",
            "revocation_endpoint": f"https://{domain}/oauth/revoke",
            "jwks_uri": f"https://{domain}/.well-known/jwks.json",
        }
    )

    expected_username = "testuser"

    def extra_settings(self):
        settings = super().extra_settings()
        settings.update(
            {
                "SOCIAL_AUTH_AUTH0_OPENIDCONNECT_DOMAIN": self.domain,
                "SOCIAL_AUTH_AUTH0_OPENIDCONNECT_USERNAME_KEY": "nickname",
            }
        )
        return settings

    def pre_complete_callback(self, start_url) -> None:
        super().pre_complete_callback(start_url)

        # Mock userinfo response with Auth0-specific fields
        responses.add(
            "GET",
            url=self.backend.userinfo_url(),
            status=200,
            body=json.dumps(
                {
                    "nickname": self.expected_username,
                    "email": "test@example.com",
                    "email_verified": True,
                    "picture": "https://example.com/avatar.jpg",
                    "locale": "en-US",
                    "sub": "auth0|123456789",
                    "name": "Test User",
                    "given_name": "Test",
                    "family_name": "User",
                }
            ),
            content_type="application/json",
        )

    def test_domain_configuration(self) -> None:
        """Test that domain-based URLs are constructed correctly"""
        self.assertEqual(
            self.backend.authorization_url(), f"https://{self.domain}/authorize"
        )
        self.assertEqual(
            self.backend.access_token_url(), f"https://{self.domain}/oauth/token"
        )

    def test_auth0_user_details(self) -> None:
        """Test Auth0-specific user detail extraction"""
        response = {
            "nickname": "testuser",
            "email": "test@example.com",
            "email_verified": True,
            "picture": "https://example.com/avatar.jpg",
            "locale": "en-US",
            "sub": "auth0|123456789",
            "name": "Test User",
            "given_name": "Test",
            "family_name": "User",
        }

        details = self.backend.get_user_details(response)

        self.assertEqual(details["username"], "testuser")
        self.assertEqual(details["email"], "test@example.com")
        self.assertTrue(details["email_verified"])
        self.assertEqual(details["picture"], "https://example.com/avatar.jpg")
        self.assertEqual(details["user_id"], "auth0|123456789")

    def test_everything_works(self) -> None:
        self.do_login()



from typing import Any

from .oauth import BaseOAuth2


class NaverOAuth2(BaseOAuth2):
    """Naver OAuth authentication backend"""

    name = "naver"
    AUTHORIZATION_URL = "https://nid.naver.com/oauth2.0/authorize"
    ACCESS_TOKEN_URL = "https://nid.naver.com/oauth2.0/token"
    EXTRA_DATA = [
        ("id", "id"),
    ]

    def get_user_id(self, details, response):
        return response.get("id")

    def get_user_details(self, response):
        """Return user details from Naver account"""
        return {
            "username": response.get("username"),
            "email": response.get("email"),
            "fullname": response.get("username"),
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        response = self.request(
            "https://openapi.naver.com/v1/nid/me",
            headers={
                "Authorization": f"Bearer {access_token}",
                "Content_Type": "text/json",
            },
        )

        data = response.json()

        return {
            "id": self._fetch(data, "id"),
            "email": self._fetch(data, "email"),
            "username": self._fetch(data, "name"),
            "nickname": self._fetch(data, "nickname"),
            "gender": self._fetch(data, "gender"),
            "age": self._fetch(data, "age"),
            "birthday": self._fetch(data, "birthday"),
            "profile_image": self._fetch(data, "profile_image"),
        }

    def auth_headers(self):
        client_id, client_secret = self.get_key_and_secret()
        return {
            "grant_type": "authorization_code",
            "code": self.data.get("code"),
            "client_id": client_id,
            "client_secret": client_secret,
        }

    def _fetch(self, data, key):
        try:
            return data["response"][key]
        except (KeyError, TypeError):
            return ""



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class PodioOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.podio.PodioOAuth2"
    user_data_url = "https://api.podio.com/user/status"
    expected_username = "user_1010101010"
    access_token_body = json.dumps(
        {
            "token_type": "bearer",
            "access_token": "11309ea9016a4ad99f1a3bcb9bc7a9d1",
            "refresh_token": "52d01df8b9ac46a4a6be1333d9f81ef2",
            "expires_in": 28800,
            "ref": {
                "type": "user",
                "id": 1010101010,
            },
        }
    )
    user_data_body = json.dumps(
        {
            "user": {
                "user_id": 1010101010,
                "activated_on": "2012-11-22 09:37:21",
                "created_on": "2012-11-21 12:23:47",
                "locale": "en_GB",
                "timezone": "Europe/Copenhagen",
                "mail": "foo@bar.com",
                "mails": [
                    {
                        "disabled": False,
                        "mail": "foobar@example.com",
                        "primary": False,
                        "verified": True,
                    },
                    {
                        "disabled": False,
                        "mail": "foo@bar.com",
                        "primary": True,
                        "verified": True,
                    },
                ],
                # more properties ...
            },
            "profile": {
                "last_seen_on": "2013-05-16 12:21:13",
                "link": "https://podio.com/users/1010101010",
                "name": "Foo Bar",
                # more properties ...
            },
            # more properties ...
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



import json
from urllib.parse import urlencode

from .oauth import OAuth1AuthUrlTestMixin, OAuth1Test


class TripitOAuth1Test(OAuth1Test, OAuth1AuthUrlTestMixin):
    backend_path = "social_core.backends.tripit.TripItOAuth"
    user_data_url = "https://api.tripit.com/v1/get/profile"
    expected_username = "foobar"
    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})
    request_token_body = urlencode(
        {
            "oauth_token_secret": "foobar-secret",
            "oauth_token": "foobar",
            "oauth_callback_confirmed": "true",
        }
    )
    user_data_content_type = "text/xml"
    user_data_body = (
        "<Response>"
        "<timestamp>1363590451</timestamp>"
        "<num_bytes>1040</num_bytes>"
        '<Profile ref="ignore-me">'
        "<ProfileEmailAddresses>"
        "<ProfileEmailAddress>"
        "<address>foobar@gmail.com</address>"
        "<is_auto_import>false</is_auto_import>"
        "<is_confirmed>true</is_confirmed>"
        "<is_primary>true</is_primary>"
        "<is_auto_inbox_eligible>"
        "true"
        "</is_auto_inbox_eligible>"
        "</ProfileEmailAddress>"
        "</ProfileEmailAddresses>"
        "<is_client>true</is_client>"
        "<is_pro>false</is_pro>"
        "<screen_name>foobar</screen_name>"
        "<public_display_name>Foo Bar</public_display_name>"
        "<profile_url>people/foobar</profile_url>"
        "<home_city>Foo, Barland</home_city>"
        "<activity_feed_url>"
        "https://www.tripit.com/feed/activities/private/"
        "ignore-this/activities.atom"
        "</activity_feed_url>"
        "<alerts_feed_url>"
        "https://www.tripit.com/feed/alerts/private/"
        "ignore-this/alerts.atom"
        "</alerts_feed_url>"
        "<ical_url>"
        "webcal://www.tripit.com/feed/ical/private/"
        "ignore-this/tripit.ics"
        "</ical_url>"
        "</Profile>"
        "</Response>"
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()


class TripitOAuth1UsernameAlternativesTest(TripitOAuth1Test):
    user_data_body = (
        "<Response>"
        "<timestamp>1363590451</timestamp>"
        "<num_bytes>1040</num_bytes>"
        '<Profile ref="ignore-me">'
        "<ProfileEmailAddresses>"
        "<ProfileEmailAddress>"
        "<address>foobar@gmail.com</address>"
        "<is_auto_import>false</is_auto_import>"
        "<is_confirmed>true</is_confirmed>"
        "<is_primary>true</is_primary>"
        "<is_auto_inbox_eligible>"
        "true"
        "</is_auto_inbox_eligible>"
        "</ProfileEmailAddress>"
        "</ProfileEmailAddresses>"
        "<is_client>true</is_client>"
        "<is_pro>false</is_pro>"
        "<screen_name>foobar</screen_name>"
        "<public_display_name>Foobar</public_display_name>"
        "<profile_url>people/foobar</profile_url>"
        "<home_city>Foo, Barland</home_city>"
        "<activity_feed_url>"
        "https://www.tripit.com/feed/activities/private/"
        "ignore-this/activities.atom"
        "</activity_feed_url>"
        "<alerts_feed_url>"
        "https://www.tripit.com/feed/alerts/private/"
        "ignore-this/alerts.atom"
        "</alerts_feed_url>"
        "<ical_url>"
        "webcal://www.tripit.com/feed/ical/private/"
        "ignore-this/tripit.ics"
        "</ical_url>"
        "</Profile>"
        "</Response>"
    )



import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class DribbbleOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.dribbble.DribbbleOAuth2"
    user_data_url = "https://api.dribbble.com/v1/user"
    expected_username = "foobar"

    access_token_body = json.dumps({"access_token": "foobar", "token_type": "bearer"})

    user_data_body = json.dumps(
        {"id": "foobar", "username": "foobar", "name": "Foo Bar"}
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()



from __future__ import annotations

from typing import TYPE_CHECKING

from .exceptions import DefaultStrategyMissingError

if TYPE_CHECKING:
    from .strategy import BaseStrategy


class Registry:
    def __init__(self) -> None:
        self._default_strategy: BaseStrategy | None = None

    def reset(self) -> None:
        self._default_strategy = None

    @property
    def default_strategy(self) -> BaseStrategy:
        if self._default_strategy is None:
            raise DefaultStrategyMissingError
        return self._default_strategy

    @default_strategy.setter
    def default_strategy(self, strategy: BaseStrategy) -> None:
        self._default_strategy = strategy


REGISTRY = Registry()
